from typing import Any, Union

import dsp
from dsp.primitives.demonstrate import Example

from .base_template import BaseTemplate


class ExperimentalAdapter(BaseTemplate):
    def query(self, example: Example, is_demo: bool = False) -> str:
        """Retrieves the input variables from the example and formats them into a query string."""
        result: list[str] = []

        # If not a demo, find the last field that doesn't have a value set in `example` and set it to ""
        # This creates the "Output:" prefix at the end of the prompt.
        if not is_demo:
            has_value = [
                field.input_variable in example
                and example[field.input_variable] is not None
                and example[field.input_variable] != ""
                for field in self.fields
            ]

            if not any(has_value):
                assert False, "No input variables found in the example"

            for i in range(1, len(has_value)):
                if has_value[i - 1] and not any(has_value[i:]):
                    example[self.fields[i].input_variable] = ""
                    break

        for field in self.fields:
            if field.input_variable in example and example[field.input_variable] is not None:
                if field.input_variable in self.format_handlers:
                    format_handler = self.format_handlers[field.input_variable]
                else:
                    def format_handler(x):
                        return str(x).strip()

                formatted_value = format_handler(example[field.input_variable])
                separator = "\n" if field.separator == " " and "\n" in formatted_value else field.separator

                result.append(f"{field.name}{separator}{formatted_value}",)

        return "\n\n".join([r for r in result if r])

    def guidelines(self, show_guidelines=True) -> str:
        """Returns the task guidelines as described in the lm prompt"""
        if (not show_guidelines) or (hasattr(dsp.settings, "show_guidelines") and not dsp.settings.show_guidelines):
            return ""

        result = "Follow the following format.\n\n"

        example = dsp.Example()
        for field in self.fields:
            example[field.input_variable] = field.description
        example.augmented = True

        result += self.query(example)
        return result

    def extract(
        self,
        example: Union[Example, dict[str, Any]],
        raw_pred: str,
    ) -> Example:
        """Extracts the answer from the LM raw prediction using the template structure

        Args:
            example (Union[Example, dict[str, Any]]): Contains the input variables that raw_pred was completed on.
            raw_pred (str): LM generated string

        Returns:
            Example: The example with the output variables filled in
        """
        example = dsp.Example(example)

        idx = 0
        while idx < len(self.fields):
            if self.fields[idx].input_variable not in example or example[self.fields[idx].input_variable] is None:
                break
            idx += 1

        import dspy

        idx = min(idx, len(self.fields) - 1)
        while raw_pred != "" and idx < len(self.fields):
            if idx < len(self.fields) - 1:
                next_field_name = "\n" + self.fields[idx + 1].name
                offset = raw_pred.find(next_field_name)

                if offset >= 0:
                    if dspy.settings.release >= 20231003:
                        example[self.fields[idx].output_variable] = raw_pred[:offset].strip().rstrip("---").strip()
                        raw_pred = raw_pred[offset + len(next_field_name) :].strip().rstrip("---").strip()
                    else:
                        field_name_parts = self.fields[idx].name.split()
                        start_pos = 0
                        for part in field_name_parts:
                            pos = raw_pred.find(part.strip())
                            if pos != -1:
                                start_pos = pos + len(part)
                            else:
                                break
                
                    example[self.fields[idx].output_variable] = raw_pred[start_pos:offset].strip().rstrip("---").strip()
                    raw_pred = raw_pred[offset + len(next_field_name) :].strip()
                    idx += 1
                else:
                    example[self.fields[idx].output_variable] = raw_pred.strip().rstrip("---").strip()

                    raw_pred = ""
                    idx += 1
                    break

            else:
                assert idx == len(self.fields) - 1, (idx, len(self.fields))

                if dspy.settings.release >= 20231003:
                    example[self.fields[idx].output_variable] = raw_pred.strip().rstrip("---").strip()
                else:
                    field_name_parts = self.fields[idx].name.split()
                    start_pos = 0
                    for part in field_name_parts:
                        pos = raw_pred.find(part.strip())
                        if pos != -1:
                            start_pos = pos + len(part)
                        else:
                            break
                    example[self.fields[idx].output_variable] = raw_pred[start_pos:].strip()

                break

        return example

    def __call__(self, example, show_guidelines=True) -> str:
        example = dsp.Example(example)
        output_fields = []
        for i in range(len(self.fields)):
            if self.fields[i].input_variable not in example:
                output_field = self.fields[i].input_variable
                if output_field not in output_fields:
                    output_fields.append(self.fields[i].name.split(':')[0])

        if hasattr(dsp.settings, "query_only") and dsp.settings.query_only:
            return self.query(example)

        # The training data should not contain the output variable
        assert self.fields[-1].input_variable not in example, f"Output variable {self.fields[-1].input_variable} should not be supplied for querying the LM."
        # del example[self.fields[-1].input_variable]

        rdemos = [
            self.query(demo, is_demo=True)
            for demo in example.demos
            if (
                (not demo.get("augmented", False))
                and (  # validate that the training example has the same primitive input var as the template
                    self.fields[-1].input_variable in demo and demo[self.fields[-1].input_variable] is not None
                )
            )
        ]

        ademos = [self.query(demo, is_demo=True) for demo in example.demos if demo.get("augmented", False)]

        # Move the rdemos to ademos if rdemo has all the fields filled in
        rdemos_ = []
        new_ademos = []
        for rdemo in rdemos:
            if all((field.name in rdemo) for field in self.fields if field.input_variable in example):
                new_ademos.append(rdemo)
            else:
                rdemos_.append(rdemo)

        ademos = new_ademos + ademos
        rdemos = rdemos_

        example["augmented"] = True

        query = self.query(example)
        parts = [self.instructions, *rdemos, self.guidelines(show_guidelines), *ademos, query,]

        prompt = "\n\n---\n\n".join([p.strip() for p in parts if p])
        prompt_ = prompt[: prompt.rfind("\n")].strip()

        s_or_not = "s" if len(output_fields) > 1 else ""
        only_or_not = "only " if len(output_fields) == 1 else ""

        prompt_ += f"\n\nPlease provide the output field{s_or_not} {', '.join(output_fields[:-1]) + (', then ' if len(output_fields) > 2 else ' then ') + output_fields[-1] if len(output_fields) > 1 else output_fields[0]}. Do so immediately, without additional content before or after, and precisely as the format above shows. Begin with {only_or_not}the field {output_fields[0]}."
        return prompt_.strip()




import json
import os
from collections import defaultdict
from typing import List, Optional, Union

import requests

import dspy
from dsp.utils import dotdict

START_SNIPPET = "<%START%>"
END_SNIPPET = "<%END%>"

def remove_snippet(s: str) -> str:
    return s.replace(START_SNIPPET, "").replace(END_SNIPPET, "")

class VectaraRM(dspy.Retrieve):
    """
    A retrieval module that uses Vectara to return the top passages for a given query.

    Assumes that a Vectara corpora have been created and populated with the following payload:
        - document: The text of the passage

    Args:
        vectara_customer_id (str): Vectara Customer ID. defaults to VECTARA_CUSTOMER_ID environment variable
        vectara_corpus_id (str): Vectara Corpus ID. defaults to VECTARA_CORPUS_ID environment variable
        vectara_api_key (str): Vectara API Key. defaults to VECTARA_API_KEY environment variable
        k (int, optional): The default number of top passages to retrieve. Defaults to 3.

    Examples:
        Below is a code snippet that shows how to use Vectara as the default retriver:
        ```python
        from vectara_client import vectaraClient

        llm = dspy.OpenAI(model="gpt-3.5-turbo")
        retriever_model = vectaraRM("<VECTARA_CUSTOMER_ID>", "<VECTARA_CORPUS_ID>", "<VECTARA_API_KEY>")
        dspy.settings.configure(lm=llm, rm=retriever_model)
        ```

        Below is a code snippet that shows how to use Vectara in the forward() function of a module
        ```python
        self.retrieve = vectaraRM("<VECTARA_CUSTOMER_ID>", "<VECTARA_CORPUS_ID>", "<VECTARA_API_KEY>", k=num_passages)
        ```
    """

    def __init__(
        self,
        vectara_customer_id: Optional[str] = None,
        vectara_corpus_id: Optional[str] = None,
        vectara_api_key: Optional[str] = None,
        k: int = 5,
    ):
        if vectara_customer_id is None:
            vectara_customer_id = os.environ.get("VECTARA_CUSTOMER_ID", "")
        if vectara_corpus_id is None:
            vectara_corpus_id = os.environ.get("VECTARA_CORPUS_ID", "")
        if vectara_api_key is None:
            vectara_api_key = os.environ.get("VECTARA_API_KEY", "")

        self._vectara_customer_id = vectara_customer_id
        self._vectara_corpus_id = vectara_corpus_id
        self._vectara_api_key = vectara_api_key
        self._n_sentences_before = self._n_sentences_after = 2
        self._vectara_timeout = 120
        super().__init__(k=k)

    def _vectara_query(
        self,
        query: str,
        limit: int = 5,
    ) -> List[str]:
        """Query Vectara index to get for top k matching passages.
        Args:
            query: query string
        """
        # If multiple corpus ids are provided (comma-separated), create a list of corpus keys
        # otherwise by default, the `split(',')` is a no-op so retains the single corpus id
        corpus_key = [
            {
                "customerId": self._vectara_customer_id,
                "corpusId": corpus_id,
                "lexicalInterpolationConfig": {"lambda": 0.025 },
            } for corpus_id in self._vectara_corpus_id.split(',')
        ]

        data = {
            "query": [
                {
                    "query": query,
                    "start": 0,
                    "numResults": limit,
                    "contextConfig": {
                        "sentencesBefore": self._n_sentences_before,
                        "sentencesAfter": self._n_sentences_after,
                        "startTag": START_SNIPPET,
                        "endTag": END_SNIPPET,
                    },
                    "corpusKey": corpus_key,
                },
            ],
        }

        headers = {
            "x-api-key": self._vectara_api_key,
            "customer-id": self._vectara_customer_id,
            "Content-Type": "application/json",
            "X-Source": "dspy",
        }

        response = requests.post(
            headers=headers,
            url="https://api.vectara.io/v1/query",
            data=json.dumps(data),
            timeout=self._vectara_timeout,
        )

        if response.status_code != 200:
            print(
                "Query failed %s",
                f"(code {response.status_code}, reason {response.reason}, details "
                f"{response.text})",
            )
            return []

        result = response.json()
        responses = result["responseSet"][0]["response"]

        res = [
            {
                "text": remove_snippet(x["text"]),
                "score": x["score"],
            } for x in responses
        ]
        return res
    
    def forward(self, query_or_queries: Union[str, List[str]], k: Optional[int]) -> dspy.Prediction:
        """Search with Vectara for self.k top passages for query

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.
            k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.
        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = (
            [query_or_queries]
            if isinstance(query_or_queries, str)
            else query_or_queries
        )
        queries = [q for q in queries if q]  # Filter empty queries        
        k = k if k is not None else self.k

        all_res = []
        limit = 3*k if len(queries) > 1 else k
        for query in queries:
            _res = self._vectara_query(query, limit=limit)
            all_res.append(_res)

        passages = defaultdict(float)
        for res_list in all_res:
            for res in res_list:
                passages[res["text"]] += res["score"]
        sorted_passages = sorted(
            passages.items(), key=lambda x: x[1], reverse=True)[:k]

        return [dotdict({"long_text": passage}) for passage, _ in sorted_passages]
    


from typing import Optional, Union, List, Any
import httpx
import logging
import os
from langfuse.client import Langfuse, StatefulTraceClient, StatefulSpanClient, StateType
from dsp.trackers.base import BaseTracker


class LangfuseTracker(BaseTracker):
    log = logging.getLogger("langfuse")

    def __init__(self, *, public_key: Optional[str] = None, secret_key: Optional[str] = None,
                 host: Optional[str] = None, debug: bool = False, stateful_client: Optional[
                Union[StatefulTraceClient, StatefulSpanClient]
            ] = None, update_stateful_client: bool = False, version: Optional[str] = None,
                 session_id: Optional[str] = None, user_id: Optional[str] = None, trace_name: Optional[str] = None,
                 release: Optional[str] = None, metadata: Optional[Any] = None, tags: Optional[List[str]] = None,
                 threads: Optional[int] = None, flush_at: Optional[int] = None, flush_interval: Optional[int] = None,
                 max_retries: Optional[int] = None, timeout: Optional[int] = None, enabled: Optional[bool] = None,
                 httpx_client: Optional[httpx.Client] = None, sdk_integration: str = "default") -> None:
        super().__init__()
        self.version = version
        self.session_id = session_id
        self.user_id = user_id
        self.trace_name = trace_name
        self.release = release
        self.metadata = metadata
        self.tags = tags

        self.root_span = None
        self.update_stateful_client = update_stateful_client
        self.langfuse = None

        prio_public_key = public_key or os.environ.get("LANGFUSE_PUBLIC_KEY")
        prio_secret_key = secret_key or os.environ.get("LANGFUSE_SECRET_KEY")
        prio_host = host or os.environ.get(
            "LANGFUSE_HOST", "https://cloud.langfuse.com"
        )

        if stateful_client and isinstance(stateful_client, StatefulTraceClient):
            self.trace = stateful_client
            self._task_manager = stateful_client.task_manager
            return

        elif stateful_client and isinstance(stateful_client, StatefulSpanClient):
            self.root_span = stateful_client
            self.trace = StatefulTraceClient(
                stateful_client.client,
                stateful_client.trace_id,
                StateType.TRACE,
                stateful_client.trace_id,
                stateful_client.task_manager,
            )
            self._task_manager = stateful_client.task_manager
            return

        args = {
            "public_key": prio_public_key,
            "secret_key": prio_secret_key,
            "host": prio_host,
            "debug": debug,
        }

        if release is not None:
            args["release"] = release
        if threads is not None:
            args["threads"] = threads
        if flush_at is not None:
            args["flush_at"] = flush_at
        if flush_interval is not None:
            args["flush_interval"] = flush_interval
        if max_retries is not None:
            args["max_retries"] = max_retries
        if timeout is not None:
            args["timeout"] = timeout
        if enabled is not None:
            args["enabled"] = enabled
        if httpx_client is not None:
            args["httpx_client"] = httpx_client
        args["sdk_integration"] = sdk_integration

        self.langfuse = Langfuse(**args)
        self.trace: Optional[StatefulTraceClient] = None
        self._task_manager = self.langfuse.task_manager

    def call(self, i, o, name=None, **kwargs):
        self.langfuse.trace(input=i, output=o, name=name, metadata=kwargs)



import dspy

from .utils import format_examples


class UnderstandTask(dspy.Signature):
    """I'll be providing you a task description. Your task is to prepare a concise, comprehensible summary that captures the broad essence and purpose of this task description. Your summary should illuminate the general objective and the type of problem being solved, offering a clear picture of what the task entails at a high level. Avoid getting into the nuances or specifics of individual datapoints, models, examples, algorithms, or any intricate technicalities. Your explanation should serve to clarify the task's overall goal and its basic premise without touching on methodologies or solutions."""

    task_description = dspy.InputField(
        prefix="Task Description:",
        desc="Description of the task.",
    )
    explanation = dspy.OutputField(
        prefix="Task Description:",
        desc="Explanation of the task.",
    )

class ExplainTask(dspy.Signature):
    """Analyze the provided set of datapoints carefully, and prepare a concise, comprehensible summary that captures the broad essence and purpose of the task these datapoints aim to address. Your summary should illuminate the general objective and the type of problem being solved, offering a clear picture of what the task entails at a high level. Avoid getting into the nuances of individual datapoints, specifics about models, examples, algorithms, or any intricate technicalities. Your explanation should serve to clarify the task's overall goal and its basic premise, without touching on methodologies or solutions."""

    examples = dspy.InputField(
        prefix="Examples Datapoints:",
        desc="List of datapoints to analyze and explain the task.",
        format=format_examples,
    )
    explanation = dspy.OutputField(
        prefix="Task Description:",
        desc="Explanation of the task.",
    )

class UpdateTaskDescriptionBasedOnFeedback(dspy.Signature):
    """Update the task description based on the feedback provided. Ensure that the revised task description incorporates the feedback to improve its overall clarity and effectiveness. Focus on enhancing the task's goal and basic premise, without delving into specific data points, models, examples, algorithms, or technical intricacies. Your explanation should aim to clarify the task's fundamental objective and purpose."""

    task_description = dspy.InputField(
        prefix="Task Description:",
        desc="Description of the task.",
    )
    feedback = dspy.InputField(
        prefix="Feedback:",
        desc="Feedback on the task description.",
    )
    updated_task_description = dspy.OutputField(
        prefix="Task Description:",
        desc="Updated description of the task.",
    )

class GetFeedbackOnGeneration(dspy.Signature):
    """Provide constructive feedback on the synthetic data generated, focusing on its quality, relevance, and diversity. Highlight any areas that require improvement and offer suggestions for enhancement. The feedback should center on the overall effectiveness of the synthetic data in aligning with the task description and knowledge seed. Avoid delving into specific data points, models, examples, algorithms, or technical intricacies. Your feedback should be critical but constructive, aiming to improve the synthetic data and the task description."""

    synthetic_data = dspy.InputField(
        prefix="Synthetic Data:",
        desc="Synthetic data generated.",
        format=format_examples,
    )
    task_description = dspy.InputField(
        prefix="Task Description:",
        desc="Description of the task the synthetic data is aligned with.",
    )
    feedback = dspy.OutputField(
        prefix="Feedback:",
        desc="Feedback on the synthetic data.",
    )

class GenerateFieldDescription(dspy.Signature):
    """Generate a concise and informative description for a given field based on the provided name and task description. This description should be no longer than 10 words and should be in simple english."""

    task_description = dspy.InputField(
        prefix="Task Description:",
        desc="Description of the task the field is an input to.",
    )
    field_name = dspy.InputField(
        prefix="Field Name:",
        desc="Name of the field to generate synthetic data for.",
    )
    field_description = dspy.OutputField(
        prefix="Field Description:",
        desc="Description of the field.",
    )

class GenerateInputFieldsData(dspy.Signature):
    """Create synthetic data using the task description and the provided knowledge seed. Your task is to generate diverse and imaginative data that aligns with the given task description and knowledge seed. You are encouraged to be creative and not limit yourself, allowing for a wide range of synthetic data that reflects the characteristics and details provided in the task description. The data should be unique and varied, showcasing originality and creativity while maintaining relevance to the task and knowledge seed.

A knowledge seed is the index of the knowledge base you have, each index represents a different knowledge base."""

    knowledge_seed = dspy.InputField(
        prefix="Knowledge Seed:",
        desc="Seed for the knowledge base search to base the inputs around.",
        format=lambda x: str(x),
    )
    task_description = dspy.InputField(
        prefix="Task Description:",
        desc="Description of the task the field is an input to.",
    )

class GenerateOutputFieldsData(dspy.Signature):
    pass


from typing import Any, Callable, Dict, List, Optional, Union

import dspy
from dspy.evaluate.evaluate import Evaluate
from dspy.primitives.example import Example
from dspy.primitives.prediction import Prediction
from dspy.primitives.program import Program
from dspy.utils.logging import logger


# TODO: Shared below are useful functions. Similar procedures are implemented
# separately and used by other DSPy teleprompters. These can be moved to shared
# locations.
def prepare_teacher(student: Program, teacher: Program = None) -> Program:
    """Prepare the teacher program with respect to the student program.
    Args:
        student: The student program.
        teacher: The teacher program. If `None`, a copy of the student program
            is used as the teacher. Defaults to `None`.
    """
    # If teacher is None, use a copy of the student program as the teacher
    if teacher is None:
        logger.info("No teacher provided. Using a copy of the student program as the teacher.")
        teacher = student.deepcopy()
    else:
        teacher = teacher.deepcopy()

    # Ensure that the student and teacher programs have the same structure
    logger.info("Ensuring that the student and teacher are are structurally equivalent.")
    student._assert_structural_equivalency(teacher)

    # Ensure that the predictors of the programs point to different objects
    logger.info("Ensuring that the student and teacher programs do not share predictors.")
    student._assert_no_shared_predictor(teacher)

    # Ensure that the LM consistency property is satisfied
    logger.info("Ensuring that the teacher program satisfies the LM consistency property.")
    teacher._assert_lm_consistency()

    # If the global LM is being used, set it to the LMs of the copied teacher
    # program predictors to to avoid handling the same edge cases later
    if dspy.settings.lm:
        teacher._set_all_predictor_lms(dspy.settings.lm)

    return teacher


def convert_to_module_level_message_data(
    data: List[Dict],
    keep_data_keys: bool = False,
    exclude_demos: bool = False,
    try_to_record_lm_kwargs: bool = False,
    program: Program = None,
) -> List[Dict]:
    """Wrapper around the function
    `build_messages_from_trace`, calling it on the "trace" field
    of each dictionary in the input data list and combiningin the results into
    a list of prompt-completion data dictionaries."""

    prompt_completion_data = []
    for data_dict in data:
        trace = data_dict["trace"]
        trace_prompt_comletion_data = build_messages_from_trace(
            trace=trace, exclude_demos=exclude_demos, try_to_record_lm_kwargs=try_to_record_lm_kwargs, program=program
        )
        for prompt_completion_dict in trace_prompt_comletion_data:
            if keep_data_keys:
                prompt_completion_dict = {**data_dict, **prompt_completion_dict}
            prompt_completion_data.append(prompt_completion_dict)
    return prompt_completion_data


def build_messages_from_trace(
    trace: List[Dict],
    exclude_demos: bool = False,
    try_to_record_lm_kwargs: bool = False,
    program: Program = None,
) -> Dict[str, List[Dict[str, Any]]]:
    messages = []
    # If the program is provided, build the predictor index to name mapping
    if program:
        pred_ind_to_name = {ind: name for ind, (name, _) in enumerate(program.named_predictors())}

    # Build the prompt-completion data

    adapter = dspy.settings.adapter or dspy.ChatAdapter()
    data = []

    # TODO: Make sure that this works for multi-stage pipelines
    for pred_ind, (pred, inputs, outputs) in enumerate(trace):
        # Get the demos from the predictor if exclude_demos is False
        demos = [] if exclude_demos else pred.demos
        messages = adapter.format(pred.signature, demos, inputs)
        messages.append(
            adapter.format_turn(signature=pred.signature, values=outputs, role="assistant", incomplete=False)
        )
        data.append(messages)

    return data


def bootstrap_data(
    program: Program,
    dataset: List[Example],
    metric: Optional[Callable[[Example, Prediction, Optional[List]], Union[bool, int, float]]] = None,
    num_threads=1,
    max_errors: int = 0,
) -> List[Dict[str, Any]]:
    """Bootstrap example, prediction, trace, example_ind, score data for the program using the dataset."""
    data = []

    # Use Evaluate to call the program have the responses cached
    cname = program.__class__.__name__
    info = f"Bootstrapping data on {len(dataset)} examples with the program {cname}, with {num_threads} threads"
    logger.info(info)
    evaluator = Evaluate(
        devset=dataset,
        num_threads=num_threads,
        display_progress=True,
        max_errors=max_errors,
        provide_traceback=True,
    )
    evaluator(program, metric=metric)

    data = []
    for example_ind, example in enumerate(dataset):
        data_dict = bootstrap_one_example(example=example, example_ind=example_ind, program=program, metric=metric)
        if data_dict is not None:
            data.append(data_dict)

    return data


def bootstrap_one_example(
    example: Any, example_ind: int, program: Program, metric: Optional[Callable] = None
) -> Dict[str, Any]:
    with dspy.context(trace=[]):
        prediction = program(**example.inputs())
        trace = dspy.settings.trace
        score = metric(example, prediction, trace) if metric else None

    data_dict = {"example": example, "prediction": prediction, "trace": trace, "example_ind": example_ind}
    if metric:
        data_dict["score"] = score

    return data_dict



import re
import time
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

import openai

from dspy.clients.finetune import (
    FinetuneJob,
    TrainingMethod,
    TrainingStatus,
    save_data,
    validate_finetune_data,
)
from dspy.utils.logging import logger

# Provider name
PROVIDER_OPENAI = "openai"


def is_openai_model(model: str) -> bool:
    """Check if the model is an OpenAI model."""
    # Filter the provider_prefix, if exists
    provider_prefix = f"{PROVIDER_OPENAI}/"
    if model.startswith(provider_prefix):
        model = model[len(provider_prefix) :]

    client = openai.OpenAI()
    valid_model_names = [model.id for model in client.models.list().data]
    # Check if the model is a base OpenAI model
    if model in valid_model_names:
        return True

    # Check if the model is a fine-tuned OpneAI model. Fine-tuned OpenAI models
    # have the prefix "ft:<BASE_MODEL_NAME>:", followed by a string specifying
    # the fine-tuned model. The following RegEx pattern is used to match the
    # base model name.
    # TODO: This part can be updated to match the actual fine-tuned model names
    # by making a call to the OpenAI API to be more exact, but this might
    # require an API key with the right permissions.
    match = re.match(r"ft:([^:]+):", model)
    if match and match.group(1) in valid_model_names:
        return True

    return False


class FinetuneJobOpenAI(FinetuneJob):
    def __init__(self, *args, **kwargs):
        self.provider_file_id = None  # TODO: Can we get this using the job_id?
        self.provider_job_id = None
        super().__init__(*args, **kwargs)

    def cancel(self):
        # Cancel the provider job
        if _does_job_exist(self.provider_job_id):
            status = _get_training_status(self.provider_job_id)
            if _is_terminal_training_status(status):
                err_msg = "Jobs that are complete cannot be canceled."
                err_msg += f" Job with ID {self.provider_job_id} is done."
                raise Exception(err_msg)
            openai.fine_tuning.jobs.cancel(self.provider_job_id)
            self.provider_job_id = None

        # Delete the provider file
        # TODO: Should there be a separate clean method?
        if self.provider_file_id is not None:
            if _does_file_exist(self.provider_file_id):
                openai.files.delete(self.provider_file_id)
            self.provider_file_id = None

        # Call the super's cancel method after the custom cancellation logic
        super().cancel()

    def status(self) -> TrainingStatus:
        status = _get_training_status(self.provider_job_id)
        return status


def finetune_openai(
    job: FinetuneJobOpenAI,
    model: str,
    train_data: List[Dict[str, Any]],
    train_kwargs: Optional[Dict[str, Any]] = None,
    train_method: TrainingMethod = TrainingMethod.SFT,
) -> str:
    train_kwargs = train_kwargs or {}
    train_method = TrainingMethod.SFT  # Note: This could be an argument; ignoring method

    # Validate train data and method
    logger.info("[Finetune] Validating the formatting of the data")
    _validate_data(train_data, train_method)
    logger.info("[Finetune] Done!")

    # Convert to the OpenAI format
    logger.info("[Finetune] Converting the data to the OpenAI format")
    # TODO: Should we use the system prompt?
    train_data = _convert_data(train_data)
    logger.info("[Finetune] Done!")

    # Save to a file
    logger.info("[Finetune] Saving the data to a file")
    data_path = save_data(train_data, provider_name=PROVIDER_OPENAI)
    logger.info("[Finetune] Done!")

    # Upload the data to the cloud
    logger.info("[Finetune] Uploading the data to the provider")
    provider_file_id = _upload_data(data_path)
    job.provider_file_id = provider_file_id
    logger.info("[Finetune] Done!")

    logger.info("[Finetune] Start remote training")
    # We utilize model and train_kwargs here
    provider_job_id = _start_remote_training(
        train_file_id=job.provider_file_id,
        model=model,
        train_kwargs=train_kwargs,
    )
    job.provider_job_id = provider_job_id
    # job.provider_job_id = "ftjob-ZdEL1mUDk0dwdDuZJQOng8Vv"
    logger.info("[Finetune] Done!")

    logger.info("[Finetune] Wait for training to complete")
    # TODO: Would it be possible to stream the logs?
    _wait_for_job(job)
    logger.info("[Finetune] Done!")

    logger.info("[Finetune] Get trained model if the run was a success")
    model = _get_trained_model(job)
    logger.info("[Finetune] Done!")

    return model


_SUPPORTED_TRAINING_METHODS = [
    TrainingMethod.SFT,
]


def _get_training_status(job_id: str) -> Union[TrainingStatus, Exception]:
    # TODO: Should this type be shared across all fine-tune clients?
    provider_status_to_training_status = {
        "validating_files": TrainingStatus.pending,
        "queued": TrainingStatus.pending,
        "running": TrainingStatus.running,
        "succeeded": TrainingStatus.succeeded,
        "failed": TrainingStatus.failed,
        "cancelled": TrainingStatus.cancelled,
    }

    # Check if there is an active job
    if job_id is None:
        logger.info("There is no active job.")
        return TrainingStatus.not_started

    err_msg = f"Job with ID {job_id} does not exist."
    assert _does_job_exist(job_id), err_msg

    # Retrieve the provider's job and report the status
    provider_job = openai.fine_tuning.jobs.retrieve(job_id)
    provider_status = provider_job.status
    status = provider_status_to_training_status[provider_status]

    return status


def _does_job_exist(job_id: str) -> bool:
    try:
        # TODO: Error handling is vague
        openai.fine_tuning.jobs.retrieve(job_id)
        return True
    except Exception:
        return False


def _does_file_exist(file_id: str) -> bool:
    try:
        # TODO: Error handling is vague
        openai.files.retrieve(file_id)
        return True
    except Exception:
        return False


def _is_terminal_training_status(status: TrainingStatus) -> bool:
    return status in [
        TrainingStatus.succeeded,
        TrainingStatus.failed,
        TrainingStatus.cancelled,
    ]


def _validate_data(data: Dict[str, str], train_method: TrainingMethod) -> Optional[Exception]:
    # Check if this train method is supported
    if train_method not in _SUPPORTED_TRAINING_METHODS:
        err_msg = f"OpenAI does not support the training method {train_method}."
        raise ValueError(err_msg)

    validate_finetune_data(data, train_method)


def _convert_data(
    data: List[Dict[str, str]],
    system_prompt: Optional[str] = None,
) -> Union[List[Dict[str, Any]], Exception]:
    # Item-wise conversion function
    def _row_converter(d):
        messages = [{"role": "user", "content": d["prompt"]}, {"role": "assistant", "content": d["completion"]}]
        if system_prompt:
            messages.insert(0, {"role": "system", "content": system_prompt})
        messages_dict = {"messages": messages}
        return messages_dict

    # Convert the data to the OpenAI format; validate the converted data
    converted_data = list(map(_row_converter, data))
    openai_data_validation(converted_data)
    return converted_data


def _upload_data(data_path: str) -> str:
    # Upload the data to the provider
    provider_file = openai.files.create(
        file=open(data_path, "rb"),
        purpose="fine-tune",
    )
    return provider_file.id


def _start_remote_training(train_file_id: str, model: id, train_kwargs: Optional[Dict[str, Any]] = None) -> str:
    train_kwargs = train_kwargs or {}
    provider_job = openai.fine_tuning.jobs.create(
        model=model,
        training_file=train_file_id,
        hyperparameters=train_kwargs,
    )
    return provider_job.id


def _wait_for_job(
    job: FinetuneJobOpenAI,
    poll_frequency: int = 60,
):
    while not _is_terminal_training_status(job.status()):
        time.sleep(poll_frequency)


def _get_trained_model(job):
    status = job.status()
    if status != TrainingStatus.succeeded:
        err_msg = f"Job status is {status}."
        err_msg += f" Must be {TrainingStatus.succeeded} to retrieve the model."
        logger.error(err_msg)
        raise Exception(err_msg)

    provider_job = openai.fine_tuning.jobs.retrieve(job.provider_job_id)
    finetuned_model = provider_job.fine_tuned_model
    return finetuned_model


# Adapted from https://cookbook.openai.com/examples/chat_finetuning_data_prep
def openai_data_validation(dataset: List[dict[str, Any]]):
    format_errors = defaultdict(int)
    for ex in dataset:
        if not isinstance(ex, dict):
            format_errors["data_type"] += 1
            continue

        messages = ex.get("messages", None)
        if not messages:
            format_errors["missing_messages_list"] += 1
            continue

        for message in messages:
            if "role" not in message or "content" not in message:
                format_errors["message_missing_key"] += 1

            if any(k not in ("role", "content", "name", "function_call", "weight") for k in message):
                format_errors["message_unrecognized_key"] += 1

            if message.get("role", None) not in ("system", "user", "assistant", "function"):
                format_errors["unrecognized_role"] += 1

            content = message.get("content", None)
            function_call = message.get("function_call", None)

            if (not content and not function_call) or not isinstance(content, str):
                format_errors["missing_content"] += 1

        if not any(message.get("role", None) == "assistant" for message in messages):
            format_errors["example_missing_assistant_message"] += 1

    # Raise an error if there are any format errors
    if format_errors:
        err_msg = "Found errors in the dataset format using the OpenAI API."
        err_msg += " Here are the number of datapoints for each error type:"
        for k, v in format_errors.items():
            err_msg += "\n    {k}: {v}"
        raise ValueError(err_msg)


def check_message_lengths(dataset: List[dict[str, Any]]) -> list[int]:
    n_missing_system = 0
    n_missing_user = 0
    n_messages = []
    convo_lens = []
    assistant_message_lens = []

    for ex in dataset:
        messages = ex["messages"]
        if not any(message["role"] == "system" for message in messages):
            n_missing_system += 1
        if not any(message["role"] == "user" for message in messages):
            n_missing_user += 1
        n_messages.append(len(messages))
        convo_lens.append(num_tokens_from_messages(messages))
        assistant_message_lens.append(num_assistant_tokens_from_messages(messages))
    n_too_long = sum([length > 16385 for length in convo_lens])

    if n_too_long > 0:
        logger.info(
            f"There are {n_too_long} examples that may be over the 16,385 token limit, they will be truncated during fine-tuning."
        )

    if n_missing_system > 0:
        logger.info(f"There are {n_missing_system} examples that are missing a system message.")

    if n_missing_user > 0:
        logger.info(f"There are {n_missing_user} examples that are missing a user message.")

    return convo_lens


def num_tokens_from_messages(messages, tokens_per_message=3, tokens_per_name=1):
    import tiktoken

    encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3
    return num_tokens


def num_assistant_tokens_from_messages(messages):
    import tiktoken

    encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 0
    for message in messages:
        if message["role"] == "assistant":
            num_tokens += len(encoding.encode(message["content"]))
    return num_tokens



from .anthropic import Claude
from .aws_models import AWSAnthropic, AWSMeta, AWSMistral, AWSModel

# Below is obsolete. It has been replaced with Bedrock class in dsp/modules/aws_providers.py
# from .bedrock import *
from .aws_providers import Bedrock, Sagemaker
from .azure_openai import AzureOpenAI
from .cache_utils import *
from .clarifai import *
from .cloudflare import *
from .cohere import *
from .colbertv2 import ColBERTv2, ColBERTv2RerankerLocal, ColBERTv2RetrieverLocal
from .databricks import *
from .dummy_lm import *
from .google import *
from .google_vertex_ai import *
from .gpt3 import *
from .groq_client import *
from .hf import HFModel
from .hf_client import Anyscale, HFClientTGI, Together
from .llama import *
from .mistral import *
from .ollama import *
from .multi_openai import MultiOpenAI
from .premai import PremAI
from .pyserini import *
from .sbert import *
from .sentence_vectorizer import *
from .snowflake import *
from .tensorrt_llm import TensorRTModel
from .watsonx import *
from .you import You



INPUT_GENERATION_TASK_WITH_EXAMPLES_SUFFIX = """\n\nI'll also be providing you some data I generated before hand, make sure the data you generate if consistent with task I provided but different from the data I provided in every way possible."""

INPUT_GENERATION_TASK_WITH_FEEDBACK_SUFFIX = "\n\nAdditionally, I'll be providing you with feedback on the data you generate, while generating the data make sure to take into account the feedback I provide and try to improve the data you generate based on the feedback I provide."


"""Module for interacting with Google Vertex AI."""

from typing import Any, Dict

import backoff
from pydantic_core import PydanticCustomError

from dsp.modules.lm import LM
from dsp.utils.settings import settings

try:
    import vertexai  # type: ignore[import-untyped]
    from vertexai.language_models import CodeGenerationModel, TextGenerationModel
    from vertexai.preview.generative_models import GenerativeModel
except ImportError:
    pass


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        f"Backing off {details['wait']:0.1f} seconds after {details['tries']} tries "
        f"calling function {details['target']} with kwargs "
        f"{details['kwargs']}",
    )


def giveup_hdlr(details):
    """wrapper function that decides when to give up on retry"""
    if "rate limits" in details.message:
        return False
    return True


class GoogleVertexAI(LM):
    """Wrapper around GoogleVertexAI's API.

    Currently supported models include `gemini-pro-1.0`.
    """

    def __init__(
        self,
        model: str = "text-bison@002",
        **kwargs,
    ):
        """
        Parameters
        ----------
        model : str
            Which pre-trained model from Google to use?
            Choices are ['gemini-1.0-pro-001', 'gemini-1.0-pro',
            'claude-3-sonnet@20240229', 'claude-3-sonnet@20240229', 'claude-3-haiku@20240307',
            'text-bison@002', 'text-bison-32k@002', 'text-bison',]
            full list at https://console.cloud.google.com/vertex-ai/model-garden
        **kwargs: dict
            Additional arguments to pass to the API provider.
        """
        super().__init__(model)
        self._is_gemini = "gemini" in model
        self._init_vertexai(kwargs)
        if "code" in model:
            model_cls = CodeGenerationModel
            self.available_args = {
                "suffix",
                "max_output_tokens",
                "temperature",
                "stop_sequences",
                "candidate_count",
            }
        elif "gemini" in model:
            model_cls = GenerativeModel
            self.available_args = {
                "max_output_tokens",
                "temperature",
                "top_k",
                "top_p",
                "stop_sequences",
                "candidate_count",
            }
        elif "text" in model:
            model_cls = TextGenerationModel
            self.available_args = {
                "max_output_tokens",
                "temperature",
                "top_k",
                "top_p",
                "stop_sequences",
                "candidate_count",
            }
        else:
            raise PydanticCustomError(
                "model",
                'model name is not valid, got "{model_name}"',
                dict(wrong_value=model),
            )
        if self._is_gemini:
            self.client = model_cls(
                model_name=model, safety_settings=kwargs.get("safety_settings"),
            )  # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
        else:
            self.client = model_cls.from_pretrained(model)
        self.provider = "googlevertexai"
        self.kwargs = {
            **self.kwargs,
            "temperature": 0.7,
            "max_output_tokens": 1024,
            "top_p": 1.0,
            "top_k": 1,
            **kwargs,
        }

    @classmethod
    def _init_vertexai(cls, values: Dict) -> None:
        vertexai.init(
            project=values.get("project"),
            location=values.get("location"),
            credentials=values.get("credentials"),
        )
        return

    def _prepare_params(
        self,
        parameters: Any,
    ) -> dict:
        stop_sequences = parameters.get("stop")
        params_mapping = {"n": "candidate_count", "max_tokens": "max_output_tokens"}
        params = {params_mapping.get(k, k): v for k, v in parameters.items()}
        params = {**self.kwargs, "stop_sequences": stop_sequences, **params}

        if self._is_gemini:
            if "candidate_count" in params and params["candidate_count"] != 1:
                print(
                    f"As of now, Gemini only supports `candidate_count == 1` (see also https://cloud.google.com/vertex-ai/generative-ai/docs/model-reference/gemini#parameters). The current value for candidate_count of {params['candidate_count']} will be overridden.",
                )
                params["candidate_count"] = 1

        return {k: params[k] for k in set(params.keys()) & self.available_args}

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs
        kwargs = self._prepare_params(raw_kwargs)
        if self._is_gemini:
            response = self.client.generate_content(
                [prompt],
                generation_config=kwargs,
            )
            history = {
                "prompt": prompt,
                "response": {
                    "prompt": prompt,
                    "choices": [
                        {
                            "text": "\n".join(v.text for v in c.content.parts),
                            "safetyAttributes": {
                                v.category: v.probability for v in c.safety_ratings
                            },
                        }
                        for c in response.candidates
                    ],
                },
                "kwargs": kwargs,
                "raw_kwargs": raw_kwargs,
            }
        else:
            response = self.client.predict(prompt, **kwargs).raw_prediction_response
            history = {
                "prompt": prompt,
                "response": {
                    "prompt": prompt,
                    "choices": [
                        {
                            "text": c["content"],
                            "safetyAttributes": c["safetyAttributes"],
                        }
                        for c in response.predictions
                    ],
                },
                "kwargs": kwargs,
                "raw_kwargs": raw_kwargs,
            }
        self.history.append(history)

        return [i["text"] for i in history["response"]["choices"]]

    @backoff.on_exception(
        backoff.expo,
        (Exception),
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of completions from Google whilst handling API errors"""
        return self.basic_request(prompt, **kwargs)

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ):
        return self.request(prompt, **kwargs)



import ast
import inspect
import re
import types
import typing
from contextlib import ExitStack, contextmanager
from copy import deepcopy
from typing import Any, Dict, Tuple, Type, Union  # noqa: UP035

from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo

import dsp
from dspy.signatures.field import InputField, OutputField, new_to_old_field


def signature_to_template(signature, adapter=None) -> dsp.Template:
    """Convert from new to legacy format."""

    adapter = adapter or dsp.Template

    return adapter(
        signature.instructions,
        **{name: new_to_old_field(field) for name, field in signature.fields.items()},
    )


def _default_instructions(cls) -> str:
    inputs_ = ", ".join([f"`{field}`" for field in cls.input_fields])
    outputs_ = ", ".join([f"`{field}`" for field in cls.output_fields])
    return f"Given the fields {inputs_}, produce the fields {outputs_}."


class SignatureMeta(type(BaseModel)):
    def __call__(cls, *args, **kwargs):  # noqa: ANN002
        if cls is Signature:
            return make_signature(*args, **kwargs)
        return super().__call__(*args, **kwargs)

    def __new__(mcs, signature_name, bases, namespace, **kwargs):  # noqa: N804
        # Set `str` as the default type for all fields
        raw_annotations = namespace.get("__annotations__", {})
        for name, field in namespace.items():
            if not isinstance(field, FieldInfo):
                continue  # Don't add types to non-field attributes
            if not name.startswith("__") and name not in raw_annotations:
                raw_annotations[name] = str
        namespace["__annotations__"] = raw_annotations

        # Let Pydantic do its thing
        cls = super().__new__(mcs, signature_name, bases, namespace, **kwargs)

        # If we don't have instructions, it might be because we are a derived generic type.
        # In that case, we should inherit the instructions from the base class.
        if cls.__doc__ is None:
            for base in bases:
                if isinstance(base, SignatureMeta):
                    doc = getattr(base, "__doc__", "")
                    if doc != "":
                        cls.__doc__ = doc

        # The more likely case is that the user has just not given us a type.
        # In that case, we should default to the input/output format.
        if cls.__doc__ is None:
            cls.__doc__ = _default_instructions(cls)

        # Ensure all fields are declared with InputField or OutputField
        cls._validate_fields()

        # Ensure all fields have a prefix
        for name, field in cls.model_fields.items():
            if "prefix" not in field.json_schema_extra:
                field.json_schema_extra["prefix"] = infer_prefix(name) + ":"
            if "desc" not in field.json_schema_extra:
                field.json_schema_extra["desc"] = f"${{{name}}}"

        return cls

    def _validate_fields(cls):
        for name, field in cls.model_fields.items():
            extra = field.json_schema_extra or {}
            field_type = extra.get("__dspy_field_type")
            if field_type not in ["input", "output"]:
                raise TypeError(
                    f"Field '{name}' in '{cls.__name__}' must be declared with "
                    "InputField or OutputField. {field.json_schema_extra=}",
                )

    @property
    def signature(cls) -> str:
        in_args = ", ".join(cls.input_fields.keys())
        out_args = ", ".join(cls.output_fields.keys())
        return f"{in_args} -> {out_args}"

    @property
    def instructions(cls) -> str:
        return inspect.cleandoc(getattr(cls, "__doc__", ""))

    @instructions.setter
    def instructions(cls, instructions: str) -> None:
        setattr(cls, "__doc__", instructions)

    def with_instructions(cls, instructions: str) -> Type["Signature"]:
        return Signature(cls.fields, instructions)

    @property
    def fields(cls) -> dict[str, FieldInfo]:
        # Make sure to give input fields before output fields
        return {**cls.input_fields, **cls.output_fields}

    def with_updated_fields(cls, name, type_=None, **kwargs) -> Type["Signature"]:
        """Update the field, name, in a new Signature type.

        Returns a new Signature type with the field, name, updated
        with fields[name].json_schema_extra[key] = value.
        """
        fields_copy = deepcopy(cls.fields)
        fields_copy[name].json_schema_extra = {
            **fields_copy[name].json_schema_extra,
            **kwargs,
        }
        if type_ is not None:
            fields_copy[name].annotation = type_
        return Signature(fields_copy, cls.instructions)

    @property
    def input_fields(cls) -> dict[str, FieldInfo]:
        return cls._get_fields_with_type("input")

    @property
    def output_fields(cls) -> dict[str, FieldInfo]:
        return cls._get_fields_with_type("output")

    def _get_fields_with_type(cls, field_type) -> dict[str, FieldInfo]:
        return {k: v for k, v in cls.model_fields.items() if v.json_schema_extra["__dspy_field_type"] == field_type}

    def prepend(cls, name, field, type_=None) -> Type["Signature"]:
        return cls.insert(0, name, field, type_)

    def append(cls, name, field, type_=None) -> Type["Signature"]:
        return cls.insert(-1, name, field, type_)

    def insert(cls, index: int, name: str, field, type_: Type = None) -> Type["Signature"]:
        # It's posisble to set the type as annotation=type in pydantic.Field(...)
        # But this may be annoying for users, so we allow them to pass the type
        if type_ is None:
            type_ = field.annotation
        if type_ is None:
            type_ = str

        input_fields = list(cls.input_fields.items())
        output_fields = list(cls.output_fields.items())

        # Choose the list to insert into based on the field type
        lst = input_fields if field.json_schema_extra["__dspy_field_type"] == "input" else output_fields
        # We support negative insert indices
        if index < 0:
            index += len(lst) + 1
        if index < 0 or index > len(lst):
            raise ValueError(f"Invalid index: {index}")
        lst.insert(index, (name, (type_, field)))

        new_fields = dict(input_fields + output_fields)
        return Signature(new_fields, cls.instructions)

    def dump_state(cls):
        state = {"instructions": cls.instructions, "fields": []}
        for field in cls.fields:
            state["fields"].append(
                {
                    "prefix": cls.fields[field].json_schema_extra["prefix"],
                    "description": cls.fields[field].json_schema_extra["desc"],
                }
            )

        return state

    def load_state(cls, state):
        signature_copy = Signature(deepcopy(cls.fields), cls.instructions)

        signature_copy.instructions = state["instructions"]
        for field, saved_field in zip(signature_copy.fields.values(), state["fields"]):
            field.json_schema_extra["prefix"] = saved_field["prefix"]
            field.json_schema_extra["desc"] = saved_field["description"]

        return signature_copy

    def equals(cls, other) -> bool:
        """Compare the JSON schema of two Pydantic models."""
        if not isinstance(other, type) or not issubclass(other, BaseModel):
            return False
        if cls.instructions != other.instructions:
            return False
        for name in cls.fields.keys() | other.fields.keys():
            if name not in other.fields or name not in cls.fields:
                return False
            # TODO: Should we compare the fields?
        return True

    def __repr__(cls):
        """Output a representation of the signature.

        Uses the form:
        Signature(question, context -> answer
            question: str = InputField(desc="..."),
            context: List[str] = InputField(desc="..."),
            answer: int = OutputField(desc="..."),
        ).
        """
        field_reprs = []
        for name, field in cls.fields.items():
            field_reprs.append(f"{name} = Field({field})")
        field_repr = "\n    ".join(field_reprs)
        return f"{cls.__name__}({cls.signature}\n    instructions={repr(cls.instructions)}\n    {field_repr}\n)"


# A signature for a predictor.
#
# You typically subclass it, like this:
# class MySignature(Signature):
#     input: str = InputField(desc="...")  # noqa: ERA001
#     output: int = OutputField(desc="...")  # noqa: ERA001
#
# You can call Signature("input1, input2 -> output1, output2") to create a new signature type.
# You can also include instructions, Signature("input -> output", "This is a test").
# But it's generally better to use the make_signature function.
#
# If you are not sure if your input is a string representation, (like "input1, input2 -> output1, output2"),
# or a signature, you can use the ensure_signature function.
#
# For compatibility with the legacy dsp format, you can use the signature_to_template function.
#


class Signature(BaseModel, metaclass=SignatureMeta):
    ""  # noqa: D419

    # Note: Don't put a docstring here, as it will become the default instructions
    # for any signature that doesn't define it's own instructions.
    pass

    @classmethod
    @contextmanager
    def replace(
        cls: "Signature",
        new_signature: "Signature",
        validate_new_signature: bool = True,
    ) -> typing.Generator[None, None, None]:
        """Replace the signature with an updated version.

        This is useful for updating the internal signatures of dspy

        Args:
            new_signature: The new signature to replace the old one with.
            validate_new_signature: Whether to validate the new signature against the old one
                to ensure that no fields are missing.
        """
        if validate_new_signature:
            for field in cls.model_fields:
                if field not in new_signature.model_fields:
                    raise ValueError(
                        f"Field '{field}' is missing from the updated signature '{new_signature.__class__}.",
                    )

        class OldSignature(cls, Signature):
            pass

        replace_fields = ["__doc__", "model_fields", "model_extra", "model_config"]
        for field in replace_fields:
            setattr(cls, field, getattr(new_signature, field))
        cls.model_rebuild(force=True)
        yield
        for field in replace_fields:
            setattr(cls, field, getattr(OldSignature, field))
        cls.model_rebuild(force=True)


@contextmanager
def update_signatures(
    signature_map: Dict[Type[Signature], Type[Signature]],
    validate_new_signature: bool = True,
) -> typing.Generator[None, None, None]:
    """Replace multiple signatures with updated versions, according to a mapping between the old and new signatures."""
    with ExitStack() as stack:
        for old_signature, new_signature in signature_map.items():
            stack.enter_context(old_signature.replace(new_signature, validate_new_signature=validate_new_signature))
        yield


def ensure_signature(signature: Union[str, Type[Signature]], instructions=None) -> Signature:
    if signature is None:
        return None
    if isinstance(signature, str):
        return Signature(signature, instructions)
    if instructions is not None:
        raise ValueError("Don't specify instructions when initializing with a Signature")
    return signature


def make_signature(
    signature: Union[str, Dict[str, Tuple[type, FieldInfo]]],
    instructions: str = None,
    signature_name: str = "StringSignature",
) -> Type[Signature]:
    """Create a new Signature type with the given fields and instructions.

    Note:
        Even though we're calling a type, we're not making an instance of the type.
        In general, instances of Signature types are not allowed to be made. The call
        syntax is provided for convenience.

    Args:
        signature: The signature format, specified as "input1, input2 -> output1, output2".
        instructions: An optional prompt for the signature.
        signature_name: An optional name for the new signature type.
    """
    fields = _parse_signature(signature) if isinstance(signature, str) else signature

    # Validate the fields, this is important because we sometimes forget the
    # slightly unintuitive syntax with tuples of (type, Field)
    fixed_fields = {}
    for name, type_field in fields.items():
        if not isinstance(name, str):
            raise ValueError(f"Field names must be strings, not {type(name)}")
        if isinstance(type_field, FieldInfo):
            type_ = type_field.annotation
            field = type_field
        else:
            if not isinstance(type_field, tuple):
                raise ValueError(f"Field values must be tuples, not {type(type_field)}")
            type_, field = type_field
        # It might be better to be explicit about the type, but it currently would break
        # program of thought and teleprompters, so we just silently default to string.
        if type_ is None:
            type_ = str
        # if not isinstance(type_, type) and not isinstance(typing.get_origin(type_), type):
        if not isinstance(type_, (type, typing._GenericAlias, types.GenericAlias)):
            raise ValueError(f"Field types must be types, not {type(type_)}")
        if not isinstance(field, FieldInfo):
            raise ValueError(f"Field values must be Field instances, not {type(field)}")
        fixed_fields[name] = (type_, field)

    # Fixing the fields shouldn't change the order
    assert list(fixed_fields.keys()) == list(fields.keys())  # noqa: S101

    # Default prompt when no instructions are provided
    if instructions is None:
        sig = Signature(signature, "")  # Simple way to parse input/output fields
        instructions = _default_instructions(sig)

    return create_model(
        signature_name,
        __base__=Signature,
        __doc__=instructions,
        **fixed_fields,
    )


def _parse_signature(signature: str) -> Tuple[Type, Field]:
    if signature.count("->") != 1:
        raise ValueError(f"Invalid signature format: '{signature}', must contain exactly one '->'.")

    inputs_str, outputs_str = signature.split("->")

    fields = {}
    for name, type_ in _parse_arg_string(inputs_str):
        fields[name] = (type_, InputField())
    for name, type_ in _parse_arg_string(outputs_str):
        fields[name] = (type_, OutputField())

    return fields


def _parse_arg_string(string: str, names=None) -> Dict[str, str]:
    args = ast.parse("def f(" + string + "): pass").body[0].args.args
    names = [arg.arg for arg in args]
    types = [str if arg.annotation is None else _parse_type_node(arg.annotation) for arg in args]
    return zip(names, types)


def _parse_type_node(node, names=None) -> Any:
    """Recursively parse an AST node representing a type annotation.

    without using structural pattern matching introduced in Python 3.10.
    """
    if names is None:
        names = typing.__dict__

    if isinstance(node, ast.Module):
        body = node.body
        if len(body) != 1:
            raise ValueError(f"Code is not syntactically valid: {node}")
        return _parse_type_node(body[0], names)

    if isinstance(node, ast.Expr):
        value = node.value
        return _parse_type_node(value, names)

    if isinstance(node, ast.Name):
        id_ = node.id
        if id_ in names:
            return names[id_]
        for type_ in [int, str, float, bool, list, tuple, dict]:
            if type_.__name__ == id_:
                return type_
        raise ValueError(f"Unknown name: {id_}")

    if isinstance(node, ast.Subscript):
        base_type = _parse_type_node(node.value, names)
        arg_type = _parse_type_node(node.slice, names)
        return base_type[arg_type]

    if isinstance(node, ast.Tuple):
        elts = node.elts
        return tuple(_parse_type_node(elt, names) for elt in elts)

    if isinstance(node, ast.Call) and node.func.id == "Field":
        keys = [kw.arg for kw in node.keywords]
        values = [kw.value.value for kw in node.keywords]
        return Field(**dict(zip(keys, values)))

    raise ValueError(f"Code is not syntactically valid: {node}")


def infer_prefix(attribute_name: str) -> str:
    """Infer a prefix from an attribute name."""
    # Convert camelCase to snake_case, but handle sequences of capital letters properly
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", attribute_name)
    intermediate_name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1)

    # Insert underscores around numbers to ensure spaces in the final output
    with_underscores_around_numbers = re.sub(
        r"([a-zA-Z])(\d)",
        r"\1_\2",
        intermediate_name,
    )
    with_underscores_around_numbers = re.sub(
        r"(\d)([a-zA-Z])",
        r"\1_\2",
        with_underscores_around_numbers,
    )

    # Convert snake_case to 'Proper Title Case', but ensure acronyms are uppercased
    words = with_underscores_around_numbers.split("_")
    title_cased_words = []
    for word in words:
        if word.isupper():
            title_cased_words.append(word)
        else:
            title_cased_words.append(word.capitalize())

    return " ".join(title_cased_words)



import os
import random
import re
import shutil
import subprocess
from typing import Literal

# from dsp.modules.adapter import TurboAdapter, DavinciAdapter, LlamaAdapter
import backoff
import requests

from dsp.modules.cache_utils import CacheMemory, NotebookCacheMemory
from dsp.modules.hf import HFModel, openai_to_hf
from dsp.utils.settings import settings

ERRORS = Exception


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


class HFClientTGI(HFModel):
    def __init__(self, model, port, url="http://future-hgx-1", http_request_kwargs=None, **kwargs):
        super().__init__(model=model, is_client=True)

        self.url = url
        self.ports = port if isinstance(port, list) else [port]
        self.http_request_kwargs = http_request_kwargs or {}

        self.headers = self.http_request_kwargs.pop('headers', {"Content-Type": "application/json"})

        self.kwargs = {
            "model": model,
            "port": port,
            "url": url,
            "temperature": 0.01,
            "max_tokens": 75,
            "top_p": 0.97,
            "n": 1,
            "stop": ["\n", "\n\n"],
            **kwargs,
        }

        # print(self.kwargs)

    def _generate(self, prompt, **kwargs):
        kwargs = {**self.kwargs, **kwargs}

        payload = {
            "inputs": prompt,
            "parameters": {
                "do_sample": kwargs["n"] > 1,
                "best_of": kwargs["n"],
                "details": kwargs["n"] > 1,
                # "max_new_tokens": kwargs.get('max_tokens', kwargs.get('max_new_tokens', 75)),
                # "stop": ["\n", "\n\n"],
                **kwargs,
            },
        }

        payload["parameters"] = openai_to_hf(**payload["parameters"])

        payload["parameters"]["temperature"] = max(
            0.1,
            payload["parameters"]["temperature"],
        )

        # print(payload['parameters'])

        # response = requests.post(self.url + "/generate", json=payload, headers=self.headers)

        response = send_hftgi_request_v01_wrapped(
            f"{self.url}:{random.Random().choice(self.ports)}" + "/generate",
            url=self.url,
            ports=tuple(self.ports),
            json=payload,
            headers=self.headers,
            **self.http_request_kwargs,
        )

        try:
            json_response = response.json()
            # completions = json_response["generated_text"]

            completions = [json_response["generated_text"]]

            if "details" in json_response and "best_of_sequences" in json_response["details"]:
                completions += [x["generated_text"] for x in json_response["details"]["best_of_sequences"]]

            response = {"prompt": prompt, "choices": [{"text": c} for c in completions]}
            return response
        except Exception:
            print("Failed to parse JSON response:", response.text)
            raise Exception("Received invalid JSON response from server")


@CacheMemory.cache(ignore=["arg"])
def send_hftgi_request_v01(arg, url, ports, **kwargs):
    return requests.post(arg, **kwargs)


# @functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache(ignore=["arg"])
def send_hftgi_request_v01_wrapped(arg, url, ports, **kwargs):
    return send_hftgi_request_v01(arg, url, ports, **kwargs)


@CacheMemory.cache
def send_hftgi_request_v00(arg, **kwargs):
    return requests.post(arg, **kwargs)


class HFClientVLLM(HFModel):
    def __init__(
        self,
        model,
        port,
        model_type: Literal["chat", "text"] = "text",
        url="http://localhost",
        http_request_kwargs=None,
        **kwargs,
    ):
        super().__init__(model=model, is_client=True)

        if isinstance(url, list):
            self.urls = url

        elif isinstance(url, str):
            self.urls = [f"{url}:{port}"]

        else:
            raise ValueError(
                f"The url provided to `HFClientVLLM` is neither a string nor a list of strings. It is of type {type(url)}."
            )

        self.urls_const = tuple(self.urls)
        self.port = port
        self.http_request_kwargs = http_request_kwargs or {}
        self.model_type = model_type
        self.headers = {"Content-Type": "application/json"}
        self.kwargs |= kwargs
        # kwargs needs to have model, port and url for the lm.copy() to work properly
        self.kwargs.update(
            {
                "port": port,
                "url": self.urls_const,
            }
        )

    def _generate(self, prompt, **kwargs):
        kwargs = {**self.kwargs, **kwargs}

        # Round robin the urls.
        url = self.urls.pop(0)
        self.urls.append(url)

        list_of_elements_to_allow = [
            "n",
            "best_of",
            "presence_penalty",
            "frequency_penalty",
            "repetition_penalty",
            "temperature",
            "top_p",
            "top_k",
            "min_p",
            "seed",
            "use_beam_search",
            "length_penalty",
            "early_stopping",
            "stop",
            "stop_token_ids",
            "include_stop_str_in_output",
            "ignore_eos",
            "max_tokens",
            "min_tokens",
            "logprobs",
            "prompt_logprobs",
            "detokenize",
            "skip_special_tokens",
            "spaces_between_special_tokens",
            "logits_processors",
            "truncate_prompt_tokens",
        ]
        req_kwargs = {k: v for k, v in kwargs.items() if k in list_of_elements_to_allow}

        if self.model_type == "chat":
            system_prompt = kwargs.get("system_prompt", None)
            messages = [{"role": "user", "content": prompt}]
            if system_prompt:
                messages.insert(0, {"role": "system", "content": system_prompt})

            payload = {
                "model": self.kwargs["model"],
                "messages": messages,
                **req_kwargs,
            }
            response = send_hfvllm_request_v01_wrapped(
                f"{url}/v1/chat/completions",
                url=self.urls_const,
                port=self.port,
                json=payload,
                headers=self.headers,
                **self.http_request_kwargs,
            )

            try:
                json_response = response.json()
                completions = json_response["choices"]
                response = {
                    "prompt": prompt,
                    "choices": [{"text": c["message"]["content"]} for c in completions],
                }
                return response

            except Exception:
                print("Failed to parse JSON response:", response.text)
                raise Exception("Received invalid JSON response from server")
        else:
            payload = {
                "model": self.kwargs["model"],
                "prompt": prompt,
                **req_kwargs,
            }

            response = send_hfvllm_request_v01_wrapped(
                f"{url}/v1/completions",
                url=self.urls_const,
                port=self.port,
                json=payload,
                headers=self.headers,
                **self.http_request_kwargs,
            )

            try:
                json_response = response.json()
                completions = json_response["choices"]
                response = {
                    "prompt": prompt,
                    "choices": [{"text": c["text"]} for c in completions],
                }
                return response

            except Exception:
                print("Failed to parse JSON response:", response.text)
                raise Exception("Received invalid JSON response from server")


@CacheMemory.cache(ignore=["arg"])
def send_hfvllm_request_v01(arg, url, port, **kwargs):
    return requests.post(arg, **kwargs)


# @functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache(ignore=["arg"])
def send_hfvllm_request_v01_wrapped(arg, url, port, **kwargs):
    return send_hftgi_request_v01(arg, url, port, **kwargs)


@CacheMemory.cache
def send_hfvllm_request_v00(arg, **kwargs):
    return requests.post(arg, **kwargs)


@CacheMemory.cache
def send_hfvllm_chat_request_v00(arg, **kwargs):
    return requests.post(arg, **kwargs)


class HFServerTGI:
    def __init__(self, user_dir):
        self.model_weights_dir = os.path.abspath(os.path.join(os.getcwd(), "text-generation-inference", user_dir))
        if not os.path.exists(self.model_weights_dir):
            os.makedirs(self.model_weights_dir)

    def close_server(self, port):
        process = subprocess.Popen(["docker", "ps"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, _ = process.communicate()
        print(stdout)
        if stdout:
            container_ids = stdout.decode().strip().split("\n")
            container_ids = container_ids[1:]
            for container_id in container_ids:
                match = re.search(r"^([a-zA-Z0-9]+)", container_id)
                if match:
                    container_id = match.group(1)
                    port_mapping = subprocess.check_output(["docker", "port", container_id]).decode().strip()
                    if f"0.0.0.0:{port}" in port_mapping:
                        subprocess.run(["docker", "stop", container_id], check=False)

    def run_server(
        self,
        port,
        model_name=None,
        model_path=None,
        env_variable=None,
        gpus="all",
        num_shard=1,
        max_input_length=4000,
        max_total_tokens=4096,
        max_best_of=100,
    ):
        self.close_server(port)
        if model_path:
            model_file_name = os.path.basename(model_path)
            link_path = os.path.join(self.model_weights_dir, model_file_name)
            shutil.copytree(model_path, link_path)
            model_name = (
                os.path.sep + os.path.basename(self.model_weights_dir) + os.path.sep + os.path.basename(model_path)
            )
        docker_command = f"docker run --gpus {gpus} --shm-size 1g -p {port}:80 -v {self.model_weights_dir}:{os.path.sep + os.path.basename(self.model_weights_dir)} -e {env_variable} ghcr.io/huggingface/text-generation-inference:1.1.0 --model-id {model_name} --num-shard {num_shard} --max-input-length {max_input_length} --max-total-tokens {max_total_tokens} --max-best-of {max_best_of}"
        print(f"Connect Command: {docker_command}")
        docker_process = subprocess.Popen(
            docker_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
        )
        connected = False
        output = []
        while True:
            line = docker_process.stdout.readline()
            if not line:
                break
            output.append(line.strip())
            if "Connected" in line:
                connected = True
                break
        if not connected:
            print("Could not connect to server. Error log:")
            for line in output:
                print(line)
            docker_process.terminate()
        docker_process.wait()


class Together(HFModel):
    def __init__(self, model, api_base="https://api.together.xyz/v1", api_key=None, **kwargs):
        super().__init__(model=model, is_client=True)
        self.session = requests.Session()
        self.api_base = os.getenv("TOGETHER_API_BASE") or api_base
        assert not self.api_base.endswith("/"), "Together base URL shouldn't end with /"
        self.token = os.getenv("TOGETHER_API_KEY") or api_key

        self.model = model

        self.use_inst_template = False
        if any(keyword in self.model.lower() for keyword in ["inst", "instruct"]):
            self.use_inst_template = True

        stop_default = "\n\n---"

        self.kwargs = {
            "model": model,
            "temperature": 0.0,
            "max_tokens": 512,
            "top_p": 1,
            "top_k": 20,
            "repetition_penalty": 1,
            "n": 1,
            "stop": stop_default if "stop" not in kwargs else kwargs["stop"],
            **kwargs,
        }

    @backoff.on_exception(
        backoff.expo,
        ERRORS,
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
    )
    def _generate(self, prompt, use_chat_api=False, **kwargs):
        kwargs = {**self.kwargs, **kwargs}

        stop = kwargs.get("stop")
        temperature = kwargs.get("temperature")
        max_tokens = kwargs.get("max_tokens", 150)
        top_p = kwargs.get("top_p", 0.7)
        top_k = kwargs.get("top_k", 50)
        repetition_penalty = kwargs.get("repetition_penalty", 1)
        prompt = f"[INST]{prompt}[/INST]" if self.use_inst_template else prompt

        if use_chat_api:
            url = f"{self.api_base}/chat/completions"
            messages = [
                {
                    "role": "system",
                    "content": "You are a helpful assistant. You must continue the user text directly without *any* additional interjections.",
                },
                {"role": "user", "content": prompt},
            ]
            body = {
                "model": self.model,
                "messages": messages,
                "temperature": temperature,
                "max_tokens": max_tokens,
                "top_p": top_p,
                "top_k": top_k,
                "repetition_penalty": repetition_penalty,
                "stop": stop,
            }
        else:
            url = f"{self.api_base}/completions"
            body = {
                "model": self.model,
                "prompt": prompt,
                "temperature": temperature,
                "max_tokens": max_tokens,
                "top_p": top_p,
                "top_k": top_k,
                "repetition_penalty": repetition_penalty,
                "stop": stop,
            }

        headers = {"Authorization": f"Bearer {self.token}"}

        try:
            with self.session.post(url, headers=headers, json=body) as resp:
                resp_json = resp.json()
                if use_chat_api:
                    completions = [resp_json.get("choices", [])[0].get("message", {}).get("content", "")]
                else:
                    completions = [resp_json.get("choices", [])[0].get("text", "")]
                response = {"prompt": prompt, "choices": [{"text": c} for c in completions]}
                return response
        except Exception as e:
            if resp_json:
                print(f"resp_json:{resp_json}")
            print(f"Failed to parse JSON response: {e}")
            raise Exception("Received invalid JSON response from server")


class Anyscale(HFModel):
    def __init__(self, model, api_base="https://api.endpoints.anyscale.com/v1", api_key=None, **kwargs):
        super().__init__(model=model, is_client=True)
        self.session = requests.Session()
        self.api_base = os.getenv("ANYSCALE_API_BASE") or api_base
        assert not self.api_base.endswith("/"), "Anyscale base URL shouldn't end with /"
        self.token = os.getenv("ANYSCALE_API_KEY") or api_key
        self.model = model
        self.kwargs = {
            "temperature": 0.0,
            "n": 1,
            **kwargs,
        }

    def _generate(self, prompt, use_chat_api=False, **kwargs):
        url = f"{self.api_base}/completions"

        kwargs = {**self.kwargs, **kwargs}

        temperature = kwargs.get("temperature")
        max_tokens = kwargs.get("max_tokens", 150)

        if use_chat_api:
            url = f"{self.api_base}/chat/completions"
            messages = [
                {
                    "role": "system",
                    "content": "You are a helpful assistant. You must continue the user text directly without *any* additional interjections.",
                },
                {"role": "user", "content": prompt},
            ]
            body = {
                "model": self.model,
                "messages": messages,
                "temperature": temperature,
                "max_tokens": max_tokens,
            }
        else:
            body = {
                "model": self.model,
                "prompt": f"[INST]{prompt}[/INST]",
                "temperature": temperature,
                "max_tokens": max_tokens,
            }

        headers = {"Authorization": f"Bearer {self.token}"}

        try:
            completions = []
            for i in range(kwargs.get("n", 1)):
                with self.session.post(url, headers=headers, json=body) as resp:
                    resp_json = resp.json()
                    if use_chat_api:
                        completions.extend([resp_json.get("choices", [])[0].get("message", {}).get("content", "")])
                    else:
                        completions.extend([resp_json.get("choices", [])[0].get("text", "")])
            response = {"prompt": prompt, "choices": [{"text": c} for c in completions]}
            return response
        except Exception as e:
            print(f"Failed to parse JSON response: {e}")
            raise Exception("Received invalid JSON response from server")


class ChatModuleClient(HFModel):
    def __init__(self, model, model_path=None, mode="interactive"):
        """
        MLC LLM Engine Chat Client

        Args:
            model (str): HF model identifier to load and use
            model_path (str, optional): Path to compiled model library (.so/.dylib)
            mode (str, optional): Mode to run the engine, can be "local", "interactive" or "server"
        """
        super().__init__(model=model, is_client=True)

        from mlc_llm.serve.engine import MLCEngine

        # MLCEngine also supports EngineConfig and different modes
        # Default model_lib=None will download the HuggingFace model to ~/.cache/mlc_llm/model_weights/hf/
        self.cm = MLCEngine(model=model, model_lib=model_path, mode=mode)

    def _generate(self, prompt, **kwargs):
        output = self.cm.chat.completions.create(messages=[{"role": "user", "content": prompt}], model=self.model)
        try:
            completions = [{"text": choice.message.content} for choice in output.choices]
            response = {"prompt": prompt, "choices": completions}
            return response
        except Exception:
            print("Failed to parse output:", response.text)
            raise Exception("Received invalid output")


class HFClientSGLang(HFModel):
    def __init__(self, model, port, url="http://localhost", **kwargs):
        super().__init__(model=model, is_client=True)
        self.url = f"{url}:{port}"
        self.headers = {"Content-Type": "application/json"}

        self.kwargs = {
            "temperature": 0.01,
            "max_tokens": 75,
            "top_p": 0.97,
            "n": 1,
            "stop": ["\n", "\n\n"],
            **kwargs,
        }

    def _generate(self, prompt, **kwargs):
        kwargs = {**self.kwargs, **kwargs}

        payload = {
            "model": kwargs.get("model", "default"),
            "prompt": prompt,
            **kwargs,
        }

        response = send_hfsglang_request_v00(
            f"{self.url}/v1/completions",
            json=payload,
            headers=self.headers,
        )

        try:
            json_response = response.json()
            completions = json_response["choices"]
            response = {
                "prompt": prompt,
                "choices": [{"text": c["text"]} for c in completions],
            }
            return response

        except Exception:
            print("Failed to parse JSON response:", response.text)
            raise Exception("Received invalid JSON response from server")


@CacheMemory.cache
def send_hfsglang_request_v00(arg, **kwargs):
    return requests.post(arg, **kwargs)



import os
import random
import time

import ujson
from datasets.fingerprint import Hasher

import dsp
from dspy.signatures.signature import signature_to_template

from .bootstrap import BootstrapFewShot

# from dspy.primitives import Example
from .teleprompt import Teleprompter

# from .vanilla import LabeledFewShot

# from dspy.evaluate.evaluate import Evaluate


if os.environ.get("DSP_NOTEBOOK_CACHEDIR"):
    training_data_directory = os.path.join(os.environ.get("DSP_NOTEBOOK_CACHEDIR"), "compiler")
    print(training_data_directory)
else:
    training_data_directory = "local_cache/compiler"


"""
TODO: Reduce and document the dependencies.

# !pip install evaluate
# !pip install tensorboardX
# !pip install transformers[torch]
# !pip install accelerate -U
# !pip install rouge_score


fewshot_teleprompter = BootstrapFewShot(metric=lambda gold, prediction, trace: gold.answer == prediction.answer,
                                        max_bootstrapped_demos=3, max_labeled_demos=16,
                                        teacher_settings=dict(lm=turbo))

fewshot = fewshot_teleprompter.compile(MyMultiHop(passages_per_hop=2), trainset=trainset)

"""


class BootstrapFinetune(Teleprompter):
    def __init__(self, metric=None, teacher_settings={}, multitask=True):
        self.metric = metric
        self.teacher_settings = teacher_settings
        self.multitask = multitask

        metric = metric or (lambda *args: True)
        self.teleprompter = BootstrapFewShot(
            metric=metric,
            max_bootstrapped_demos=999999,
            max_labeled_demos=0,  # FIXME: TODO: Make this zero? or param, with default as 16 or 0?
            teacher_settings=teacher_settings,
        )

        if not os.path.exists(training_data_directory):
            os.makedirs(training_data_directory)

    def compile(
        self,
        student,
        *,
        teacher=None,
        trainset,
        valset=None,
        target="t5-large",
        bsize=12,
        accumsteps=1,
        lr=5e-5,
        epochs=1,
        bf16=False,
        int8=False,
        peft=False,
        path_prefix=None,
    ):
        # It's usually better to supply a few-shot teacher, rather than uncompiled module (the student).
        if teacher is None:
            print(
                "WARNING: Using a vanilla teacher. "
                "Are you sure you want to use BootstrapFinetune without a compiled teacher?",
            )

        teachers = teacher if isinstance(teacher, list) else [teacher]
        finetune_data = {}

        for teacher in teachers:
            # Dummy compilation to get bootstraps.
            compiled = self.teleprompter.compile(student, teacher=teacher, trainset=trainset)
            multitask = self.multitask

            # Prepare finetune <prompt, completion> pairs.
            for name, predictor in compiled.named_predictors():
                name_ = "all" if multitask else name
                finetune_data[name_] = [] if name_ not in finetune_data else finetune_data[name_]

                for demo in predictor.demos:
                    demo = dict(demo)

                    # TODO: FIXME: generalize.
                    template = signature_to_template(predictor.signature)
                    completion = demo.pop(template.fields[-1].output_variable)
                    prompt = template.query(dsp.Example(demos=[], **demo)).strip()

                    finetune_data[name_].append(dict(prompt=prompt, completion=completion))

        for name_ in finetune_data:
            random.Random(0).shuffle(finetune_data[name_])
            print(name_, len(finetune_data[name_]))

        #
        # Dump as files.
        #
        finetune_paths = {}

        for name in finetune_data:
            data = finetune_data[name]
            hashed_name = name + "." + Hasher.hash(data)
            output_path = os.path.join(training_data_directory, f"{hashed_name}.jsonl")
            print(output_path)

            with open(output_path, "w") as f:
                for line in data:
                    f.write(ujson.dumps(line) + "\n")

            finetune_paths[name] = output_path

        #
        # Train!
        #
        import string

        compiler_config = {
            "save": "".join(
                random.Random(time.time()).choices(string.ascii_uppercase + string.digits, k=13),
            ),  # https://stackoverflow.com/a/2257449/1493011
            "peft": peft,
            "fp16": False,
            "bf16": bf16,
            "int8": int8,
            "fid": False,
            "rationale": False,
            "batch_size": bsize,
            "epochs": epochs,
            "gradient_accumulation_steps": accumsteps,  # 2,
            "lr": lr,
        }

        compiler_config["save"] = (
            os.path.join(path_prefix, compiler_config["save"]) if path_prefix else compiler_config["save"]
        )

        from dsp.modules.finetuning import finetune_hf

        target = target
        finetune_models = {}

        for name in finetune_data:
            training_data_path = finetune_paths[name]
            compiler_config_ = dict(compiler_config)
            compiler_config_["save"] = compiler_config["save"] + "." + name
            best_ckpt_path = finetune_hf(training_data_path, target, compiler_config_)

            print(f"#> Best checkpoint path: {best_ckpt_path} for {name}")
            finetune_models[name] = dsp.HFModel(model=target, checkpoint=best_ckpt_path)  # best_ckpt_path

        #
        # Set the LMs to the finetuned ones, per module
        #
        compiled2 = compiled.reset_copy()

        assert len(compiled.named_predictors()) == len(compiled2.named_predictors())

        for (name, predictor), (name2, predictor2) in zip(compiled.named_predictors(), compiled2.named_predictors()):
            assert name == name2
            name = "all" if multitask else name

            # TODO: FIXME: When we assign .lm, the Predict.forward will also set only_query=True.
            # This is correct for here but we may want to make it more explicitly restricted to finetuned models.
            print(f"Assigning the LM of predictor {name}.")

            predictor2.lm = finetune_models[name]
            assert predictor2.demos == []

        return compiled2



from typing import Any, Literal
from dsp.modules.lm import LM

class LlamaCpp(LM):
    def __init__(
            self,
            model: str,  # "llama" or the actual model name
            llama_model: Any = None,
            model_type: Literal["chat", "text"] = None,
            **kwargs,
    ):
        super().__init__(model)

        try:
            import llama_cpp
            from llama_cpp import Llama
        except ImportError as exc:
            raise ModuleNotFoundError(
                """You need to install the llama_cpp library to use gguf models.
                CPU - pip install llama-cpp-python
                CUDA - pip install llama-cpp-python --extra-index-url https://abetlen.github.io/llama-cpp-python/whl/<cuda-version> e.g. cu121 for CUDA 12.1
                METAL(Mac) - pip install llama-cpp-python --extra-index-url https://abetlen.github.io/llama-cpp-python/whl/metal
                others: https://pypi.org/project/llama-cpp-python/""",
            ) from exc

        default_model_type = "text"
        self.model_type = model_type if model_type else default_model_type
        self.provider = "llama"

        self.kwargs = {
            "temperature": 0.0,
            "max_tokens": 150,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            **kwargs,
        }  # TODO: add kwargs above for </s>

        self.loaded_model = llama_model
        self.history: list[dict[str, Any]] = []

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs

        kwargs = {**self.kwargs, **kwargs}
        if "n" in kwargs:
            del kwargs["n"]

        if self.model_type == "chat":
            kwargs["messages"] = [{"role": "user", "content": prompt}]
            response = self.loaded_model.create_chat_completion(**kwargs)

        else:
            kwargs["prompt"] = prompt
            response = self.loaded_model.create_completion(**kwargs)

        history = {
            "prompt": prompt,
            "response": response,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)
        return response

    def request(self, prompt: str, **kwargs):
        if "model_type" in kwargs:
            del kwargs["model_type"]

        return self.basic_request(prompt, **kwargs)

    def _get_choice_text(self, choice) -> str:
        if self.model_type == "chat":
            return choice["message"]["content"]
        return choice["text"]

    def __call__(
            self,
            prompt: str,
            only_completed: bool = True,
            return_sorted: bool = False,
            **kwargs,
    ) -> list[dict[str, Any]]:
        assert only_completed, "for now"
        assert return_sorted is False, "for now"

        response = self.request(prompt, **kwargs)
        choices = response["choices"]

        completed_choices = [c for c in choices if c["finish_reason"] != "length"]

        if only_completed and len(completed_choices):
            choices = completed_choices

        completions = [self._get_choice_text(c) for c in choices]

        return completions



import re
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional

from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.llms.generic_utils import (
    prompt_to_messages,
)
from llama_index.core.base.llms.types import ChatMessage
from llama_index.core.base.query_pipeline.query import InputKeys, OutputKeys, QueryComponent
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.prompts import BasePromptTemplate, PromptTemplate
from llama_index.core.query_pipeline import QueryPipeline

import dsp
import dspy
from dspy import Predict
from dspy.signatures.field import InputField, OutputField
from dspy.signatures.signature import ensure_signature, make_signature, signature_to_template


def get_formatted_template(predict_module: Predict, kwargs: Dict[str, Any]) -> str:
    """Get formatted template from predict module."""
    # Extract the three privileged keyword arguments.
    signature = ensure_signature(predict_module.signature)
    demos = predict_module.demos

    # All of the other kwargs are presumed to fit a prefix of the signature.
    # That is, they are input variables for the bottom most generation, so
    # we place them inside the input - x - together with the demos.
    x = dsp.Example(demos=demos, **kwargs)

    # Switch to legacy format for dsp.generate
    template = signature_to_template(signature)

    return template(x)


def replace_placeholder(text: str) -> str:
    # Use a regular expression to find and replace ${...} with ${{...}}
    return re.sub(r'\$\{([^\{\}]*)\}', r'${{\1}}', text)


def _input_keys_from_template(template: dsp.Template) -> InputKeys:
    """Get input keys from template."""
    # get only fields that are marked OldInputField and NOT OldOutputField
    # template_vars = list(template.kwargs.keys())
    return [
        k for k, v in template.kwargs.items() if isinstance(v, dspy.signatures.OldInputField)
    ]

def _output_keys_from_template(template: dsp.Template) -> InputKeys:
    """Get output keys from template."""
    # get only fields that are marked OldOutputField and NOT OldInputField
    # template_vars = list(template.kwargs.keys())
    return [
        k for k, v in template.kwargs.items() if isinstance(v, dspy.signatures.OldOutputField)
    ]


class DSPyPromptTemplate(BasePromptTemplate):
    """A prompt template for DSPy.

    Takes in a predict module from DSPy (whether unoptimized or optimized),
    and extracts the relevant prompt template from it given the input.
    
    """

    predict_module: Predict

    def __init__(
        self,
        predict_module: Predict,
        metadata: Optional[Dict[str, Any]] = None,
        template_var_mappings: Optional[Dict[str, Any]] = None,
        function_mappings: Optional[Dict[str, Callable]] = None,
        **kwargs: Any,
    ) -> None:
        template = signature_to_template(predict_module.signature)
        template_vars = _input_keys_from_template(template)
        # print(f"TEMPLATE VARS: {template_vars}")
        # raise Exception

        super().__init__(
            predict_module=predict_module,
            metadata=metadata or {},
            template_vars=template_vars,
            kwargs=kwargs,
            template_var_mappings=template_var_mappings,
            function_mappings=function_mappings,
        )
    
    def partial_format(self, **kwargs: Any) -> "BasePromptTemplate":
        """Returns a new prompt template with the provided kwargs."""
        # NOTE: this is a copy of the implementation in `PromptTemplate`
        output_parser = self.output_parser
        self.output_parser = None

        # get function and fixed kwargs, and add that to a copy
        # of the current prompt object
        prompt = deepcopy(self)
        prompt.kwargs.update(kwargs)

        # NOTE: put the output parser back
        prompt.output_parser = output_parser
        self.output_parser = output_parser
        return prompt

    def format(self, llm: Optional[BaseLLM] = None, **kwargs: Any) -> str:
        """Formats the prompt template."""
        mapped_kwargs = self._map_all_vars(kwargs)
        return get_formatted_template(self.predict_module, mapped_kwargs)

    def format_messages(
        self, llm: Optional[BaseLLM] = None, **kwargs: Any,
    ) -> List[ChatMessage]:
        """Formats the prompt template into chat messages."""
        del llm  # unused
        prompt = self.format(**kwargs)
        return prompt_to_messages(prompt)

    def get_template(self, llm: Optional[BaseLLM] = None) -> str:
        """Get template."""
        # get kwarg templates
        kwarg_tmpl_map = {k: "{k}" for k in self.template_vars}

        # get "raw" template with all the values filled in with {var_name} 
        template0 = get_formatted_template(self.predict_module, kwarg_tmpl_map)
        # HACK: there are special 'format' variables of the form ${var_name} that are meant to 
        # prompt the LLM, but we do NOT want to replace with actual prompt variable values. 
        # Replace those with double brackets
        template1 = replace_placeholder(template0)

        return template1


# copied from langchain.py
class Template2Signature(dspy.Signature):
    """You are a processor for prompts. I will give you a prompt template (Python f-string) for an arbitrary task for other LMs.
Your job is to prepare three modular pieces: (i) any essential task instructions or guidelines, (ii) a list of variable names for inputs, (iv) the variable name for output."""

    template = dspy.InputField(format=lambda x: f"```\n\n{x.strip()}\n\n```\n\nLet's now prepare three modular pieces.")
    essential_instructions = dspy.OutputField()
    input_keys = dspy.OutputField(desc='comma-separated list of valid variable names')
    output_key = dspy.OutputField(desc='a valid variable name')


def build_signature(prompt: PromptTemplate) -> dspy.Signature:
    """Attempt to build signature from prompt."""
    # TODO: allow plugging in any llamaindex LLM 
    gpt4T = dspy.OpenAI(model='gpt-4-1106-preview', max_tokens=4000, model_type='chat')

    with dspy.context(lm=gpt4T): 
        parts = dspy.Predict(Template2Signature)(template=prompt.template)

    inputs = {k.strip(): InputField() for k in parts.input_keys.split(',')}
    outputs = {k.strip(): OutputField() for k in parts.output_key.split(',')}

    # dynamically create a pydantic model that subclasses dspy.Signature
    fields = {
        k: (str, v) for k, v in {**inputs, **outputs}.items()
    }
    signature = make_signature(fields, parts.essential_instructions)
    return signature
    

class DSPyComponent(QueryComponent):
    """DSPy Query Component. 
    
    Can take in either a predict module directly.
    TODO: add ability to translate from an existing prompt template / LLM.
    
    """
    predict_module: dspy.Predict
    predict_template: dsp.Template

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        predict_module: dspy.Predict,
    ) -> None:
        """Initialize."""
        return super().__init__(
            predict_module=predict_module,
            predict_template=signature_to_template(predict_module.signature),
        )

    @classmethod
    def from_prompt(
        cls,
        prompt_template: BasePromptTemplate,
        # llm: BaseLLM,
    ) -> "DSPyComponent":
        """Initialize from prompt template.

        LLM is a TODO - currently use DSPy LLM classes.
        
        """
        signature = build_signature(prompt_template)
        predict_module = Predict(signature)
        return cls(predict_module=predict_module)
    
    def set_callback_manager(self, callback_manager: CallbackManager) -> None:
        """Set callback manager."""
        # TODO: implement
        pass

    def _validate_component_inputs(self, input: Dict[str, Any]) -> Dict[str, Any]:
        """Validate component inputs during run_component."""
        return input

    def _run_component(self, **kwargs: Any) -> Dict:
        """Run component."""
        prediction = self.predict_module(**kwargs)
        return {
            k: getattr(prediction, k) for k in self.output_keys.required_keys
        }

    async def _arun_component(self, **kwargs: Any) -> Any:
        """Run component (async)."""
        # TODO: no async predict module yet
        return self._run_component(**kwargs)

    @property
    def input_keys(self) -> InputKeys:
        """Input keys."""
        input_keys = _input_keys_from_template(self.predict_template)
        return InputKeys.from_keys(input_keys)

    @property
    def output_keys(self) -> OutputKeys:
        """Output keys."""
        output_keys = _output_keys_from_template(self.predict_template)
        return OutputKeys.from_keys(output_keys)


class LlamaIndexModule(dspy.Module):
    """A module for LlamaIndex.

    Wraps a QueryPipeline and exposes it as a dspy module for optimization.
    
    """

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, query_pipeline: QueryPipeline) -> None:
        """Initialize."""
        super().__init__()
        self.query_pipeline = query_pipeline
        self.predict_modules = []
        for module in query_pipeline.module_dict.values():
            if isinstance(module, DSPyComponent):
                self.predict_modules.append(module.predict_module)
        

    def forward(self, **kwargs: Any) -> Dict[str, Any]:
        """Forward."""
        output_dict = self.query_pipeline.run(**kwargs, return_values_direct=False)
        return dspy.Prediction(**output_dict)
    



import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import litellm
import ujson
from litellm.caching import Cache

from dspy.clients.finetune import FinetuneJob, TrainingMethod
from dspy.clients.lm_finetune_utils import execute_finetune_job, get_provider_finetune_job_class
from dspy.utils.callback import BaseCallback, with_callbacks
from dspy.utils.logging import logger

DISK_CACHE_DIR = os.environ.get("DSPY_CACHEDIR") or os.path.join(Path.home(), ".dspy_cache")
litellm.cache = Cache(disk_cache_dir=DISK_CACHE_DIR, type="disk")
litellm.telemetry = False

if "LITELLM_LOCAL_MODEL_COST_MAP" not in os.environ:
    os.environ["LITELLM_LOCAL_MODEL_COST_MAP"] = "True"

GLOBAL_HISTORY = []


class LM:
    """
    A language model supporting chat or text completion requests for use with DSPy modules.
    """

    def __init__(
        self,
        model: str,
        model_type: Literal["chat", "text"] = "chat",
        temperature: float = 0.0,
        max_tokens: int = 1000,
        cache: bool = True,
        launch_kwargs: Optional[Dict[str, Any]] = None,
        callbacks: Optional[List[BaseCallback]] = None,
        num_retries: int = 3,
        **kwargs,
    ):
        """
        Create a new language model instance for use with DSPy modules and programs.

        Args:
            model: The model to use. This should be a string of the form ``"llm_provider/llm_name"``
                   supported by LiteLLM. For example, ``"openai/gpt-4o"``.
            model_type: The type of the model, either ``"chat"`` or ``"text"``.
            temperature: The sampling temperature to use when generating responses.
            max_tokens: The maximum number of tokens to generate per response.
            cache: Whether to cache the model responses for reuse to improve performance
                   and reduce costs.
            callbacks: A list of callback functions to run before and after each request.
            num_retries: The number of times to retry a request if it fails transiently due to
                         network error, rate limiting, etc. Requests are retried with exponential
                         backoff.
        """
        # Remember to update LM.copy() if you modify the constructor!
        self.model = model
        self.model_type = model_type
        self.cache = cache
        self.launch_kwargs = launch_kwargs or {}
        self.kwargs = dict(temperature=temperature, max_tokens=max_tokens, **kwargs)
        self.history = []
        self.callbacks = callbacks or []
        self.num_retries = num_retries

        # TODO: Arbitrary model strings could include the substring "o1-". We
        # should find a more robust way to check for the "o1-" family models.
        if "o1-" in model:
            assert (
                max_tokens >= 5000 and temperature == 1.0
            ), "OpenAI's o1-* models require passing temperature=1.0 and max_tokens >= 5000 to `dspy.LM(...)`"

    @with_callbacks
    def __call__(self, prompt=None, messages=None, **kwargs):
        # Build the request.
        cache = kwargs.pop("cache", self.cache)
        messages = messages or [{"role": "user", "content": prompt}]
        kwargs = {**self.kwargs, **kwargs}

        # Make the request and handle LRU & disk caching.
        if self.model_type == "chat":
            completion = cached_litellm_completion if cache else litellm_completion
        else:
            completion = cached_litellm_text_completion if cache else litellm_text_completion

        response = completion(
            request=ujson.dumps(dict(model=self.model, messages=messages, **kwargs)),
            num_retries=self.num_retries,
        )
        outputs = [c.message.content if hasattr(c, "message") else c["text"] for c in response["choices"]]

        # Logging, with removed api key & where `cost` is None on cache hit.
        kwargs = {k: v for k, v in kwargs.items() if not k.startswith("api_")}
        entry = dict(prompt=prompt, messages=messages, kwargs=kwargs, response=response)
        entry = dict(**entry, outputs=outputs, usage=dict(response["usage"]))
        entry = dict(**entry, cost=response.get("_hidden_params", {}).get("response_cost"))
        entry = dict(
            **entry,
            timestamp=datetime.now().isoformat(),
            uuid=str(uuid.uuid4()),
            model=self.model,
            model_type=self.model_type,
        )
        self.history.append(entry)
        GLOBAL_HISTORY.append(entry)

        return outputs

    def inspect_history(self, n: int = 1):
        _inspect_history(self.history, n)

    def launch(self):
        """Send a request to the provider to launch the model, if needed."""
        msg = f"`launch()` is called for the auto-launched model {self.model}"
        msg += " -- no action is taken!"
        logger.info(msg)

    def kill(self):
        """Send a request to the provider to kill the model, if needed."""
        msg = f"`kill()` is called for the auto-launched model {self.model}"
        msg += " -- no action is taken!"
        logger.info(msg)

    def finetune(
        self,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        train_method: TrainingMethod = TrainingMethod.SFT,
        provider: str = "openai",
        cache_finetune: bool = True,
    ) -> FinetuneJob:
        """Start model fine-tuning, if supported."""
        from dspy import settings as settings

        err = "Fine-tuning is an experimental feature."
        err += " Set `dspy.settings.experimental` to `True` to use it."
        assert settings.experimental, err

        FinetuneJobClass = get_provider_finetune_job_class(provider=provider)
        finetune_job = FinetuneJobClass(
            model=self.model,
            train_data=train_data,
            train_kwargs=train_kwargs,
            train_method=train_method,
            provider=provider,
        )

        executor = ThreadPoolExecutor(max_workers=1)
        executor.submit(execute_finetune_job, finetune_job, lm=self, cache_finetune=cache_finetune)
        executor.shutdown(wait=False)

        return finetune_job

    def copy(self, **kwargs):
        """Returns a copy of the language model with possibly updated parameters."""

        import copy

        new_instance = copy.deepcopy(self)
        new_instance.history = []

        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(new_instance, key, value)
            if (key in self.kwargs) or (not hasattr(self, key)):
                new_instance.kwargs[key] = value

        return new_instance


@functools.lru_cache(maxsize=None)
def cached_litellm_completion(request, num_retries: int):
    return litellm_completion(
        request,
        cache={"no-cache": False, "no-store": False},
        num_retries=num_retries,
    )


def litellm_completion(request, num_retries: int, cache={"no-cache": True, "no-store": True}):
    kwargs = ujson.loads(request)
    return litellm.completion(
        num_retries=num_retries,
        cache=cache,
        **kwargs,
    )


@functools.lru_cache(maxsize=None)
def cached_litellm_text_completion(request, num_retries: int):
    return litellm_text_completion(
        request,
        num_retries=num_retries,
        cache={"no-cache": False, "no-store": False},
    )


def litellm_text_completion(request, num_retries: int, cache={"no-cache": True, "no-store": True}):
    kwargs = ujson.loads(request)

    # Extract the provider and model from the model string.
    # TODO: Not all the models are in the format of "provider/model"
    model = kwargs.pop("model").split("/", 1)
    provider, model = model[0] if len(model) > 1 else "openai", model[-1]

    # Use the API key and base from the kwargs, or from the environment.
    api_key = kwargs.pop("api_key", None) or os.getenv(f"{provider}_API_KEY")
    api_base = kwargs.pop("api_base", None) or os.getenv(f"{provider}_API_BASE")

    # Build the prompt from the messages.
    prompt = "\n\n".join([x["content"] for x in kwargs.pop("messages")] + ["BEGIN RESPONSE:"])

    return litellm.text_completion(
        cache=cache,
        model=f"text-completion-openai/{model}",
        api_key=api_key,
        api_base=api_base,
        prompt=prompt,
        num_retries=num_retries,
        **kwargs,
    )


def _green(text: str, end: str = "\n"):
    return "\x1b[32m" + str(text).lstrip() + "\x1b[0m" + end


def _red(text: str, end: str = "\n"):
    return "\x1b[31m" + str(text) + "\x1b[0m" + end


def _inspect_history(history, n: int = 1):
    """Prints the last n prompts and their completions."""

    for item in history[-n:]:
        messages = item["messages"] or [{"role": "user", "content": item["prompt"]}]
        outputs = item["outputs"]
        timestamp = item.get("timestamp", "Unknown time")

        print("\n\n\n")
        print("\x1b[34m" + f"[{timestamp}]" + "\x1b[0m" + "\n")

        for msg in messages:
            print(_red(f"{msg['role'].capitalize()} message:"))
            print(msg["content"].strip())
            print("\n")

        print(_red("Response:"))
        print(_green(outputs[0].strip()))

        if len(outputs) > 1:
            choices_text = f" \t (and {len(outputs)-1} other completions)"
            print(_red(choices_text, end=""))

    print("\n\n\n")



import re
import ast
import json
import enum
import inspect
import pydantic
import textwrap

from pydantic import TypeAdapter
from collections.abc import Mapping
from pydantic.fields import FieldInfo
from typing import Any, Dict, KeysView, List, Literal, NamedTuple, get_args, get_origin

from dspy.adapters.base import Adapter
from ..signatures.field import OutputField
from ..signatures.signature import SignatureMeta
from ..signatures.utils import get_dspy_field_type

field_header_pattern = re.compile(r"\[\[ ## (\w+) ## \]\]")


class FieldInfoWithName(NamedTuple):
    name: str
    info: FieldInfo


# Built-in field indicating that a chat turn has been completed.
BuiltInCompletedOutputFieldInfo = FieldInfoWithName(name="completed", info=OutputField())


class ChatAdapter(Adapter):
    def __init__(self):
        pass

    def format(self, signature, demos, inputs):
        messages = []

        # Extract demos where some of the output_fields are not filled in.
        incomplete_demos = [demo for demo in demos if not all(k in demo for k in signature.fields)]
        complete_demos = [demo for demo in demos if demo not in incomplete_demos]
        incomplete_demos = [
            demo
            for demo in incomplete_demos
            if any(k in demo for k in signature.input_fields) and any(k in demo for k in signature.output_fields)
        ]

        demos = incomplete_demos + complete_demos

        messages.append({"role": "system", "content": prepare_instructions(signature)})

        for demo in demos:
            messages.append(format_turn(signature, demo, role="user", incomplete=demo in incomplete_demos))
            messages.append(format_turn(signature, demo, role="assistant", incomplete=demo in incomplete_demos))

        messages.append(format_turn(signature, inputs, role="user"))

        return messages

    def parse(self, signature, completion, _parse_values=True):
        sections = [(None, [])]

        for line in completion.splitlines():
            match = field_header_pattern.match(line.strip())
            if match:
                sections.append((match.group(1), []))
            else:
                sections[-1][1].append(line)

        sections = [(k, "\n".join(v).strip()) for k, v in sections]

        fields = {}
        for k, v in sections:
            if (k not in fields) and (k in signature.output_fields):
                try:
                    fields[k] = parse_value(v, signature.output_fields[k].annotation) if _parse_values else v
                except Exception as e:
                    raise ValueError(
                        f"Error parsing field {k}: {e}.\n\n\t\tOn attempting to parse the value\n```\n{v}\n```"
                    )

        if fields.keys() != signature.output_fields.keys():
            raise ValueError(f"Expected {signature.output_fields.keys()} but got {fields.keys()}")

        return fields

    def format_turn(self, signature, values, role, incomplete=False):
        return format_turn(signature, values, role, incomplete)

    def format_fields(self, signature, values, role):
        fields_with_values = {
            FieldInfoWithName(name=field_name, info=field_info): values.get(
                field_name, "Not supplied for this particular example."
            )
            for field_name, field_info in signature.fields.items()
            if field_name in values
        }

        return format_fields(fields_with_values)
        
        

def format_blob(blob):
    if "\n" not in blob and "«" not in blob and "»" not in blob:
        return f"«{blob}»"

    modified_blob = blob.replace("\n", "\n    ")
    return f"«««\n    {modified_blob}\n»»»"


def format_input_list_field_value(value: List[Any]) -> str:
    """
    Formats the value of an input field of type List[Any].

    Args:
      value: The value of the list-type input field.
    Returns:
      A string representation of the input field's list value.
    """
    if len(value) == 0:
        return "N/A"
    if len(value) == 1:
        return format_blob(value[0])

    return "\n".join([f"[{idx+1}] {format_blob(txt)}" for idx, txt in enumerate(value)])


def _serialize_for_json(value):
    if isinstance(value, pydantic.BaseModel):
        return value.model_dump()
    elif isinstance(value, list):
        return [_serialize_for_json(item) for item in value]
    elif isinstance(value, dict):
        return {key: _serialize_for_json(val) for key, val in value.items()}
    else:
        return value

def _format_field_value(field_info: FieldInfo, value: Any) -> str:
    """
    Formats the value of the specified field according to the field's DSPy type (input or output),
    annotation (e.g. str, int, etc.), and the type of the value itself.

    Args:
      field_info: Information about the field, including its DSPy field type and annotation.
      value: The value of the field.
    Returns:
      The formatted value of the field, represented as a string.
    """

    if isinstance(value, list) and field_info.annotation is str:
        # If the field has no special type requirements, format it as a nice numbere list for the LM.
        return format_input_list_field_value(value)
    elif isinstance(value, pydantic.BaseModel) or isinstance(value, dict) or isinstance(value, list):
        return json.dumps(_serialize_for_json(value))
    else:
        return str(value)



def format_fields(fields_with_values: Dict[FieldInfoWithName, Any]) -> str:
    """
    Formats the values of the specified fields according to the field's DSPy type (input or output),
    annotation (e.g. str, int, etc.), and the type of the value itself. Joins the formatted values
    into a single string, which is is a multiline string if there are multiple fields.

    Args:
      fields_with_values: A dictionary mapping information about a field to its corresponding
                          value.
    Returns:
      The joined formatted values of the fields, represented as a string.
    """
    output = []
    for field, field_value in fields_with_values.items():
        formatted_field_value = _format_field_value(field_info=field.info, value=field_value)
        output.append(f"[[ ## {field.name} ## ]]\n{formatted_field_value}")

    return "\n\n".join(output).strip()


def parse_value(value, annotation):
    if annotation is str:
        return str(value)
    
    parsed_value = value

    if isinstance(annotation, enum.EnumMeta):
        parsed_value = annotation[value]
    elif isinstance(value, str):
        try:
            parsed_value = json.loads(value)
        except json.JSONDecodeError:
            try:
                parsed_value = ast.literal_eval(value)
            except (ValueError, SyntaxError):
                parsed_value = value
    
    return TypeAdapter(annotation).validate_python(parsed_value)


def format_turn(signature: SignatureMeta, values: Dict[str, Any], role, incomplete=False) -> Dict[str, str]:
    """
    Constructs a new message ("turn") to append to a chat thread. The message is carefully formatted
    so that it can instruct an LLM to generate responses conforming to the specified DSPy signature.

    Args:
      signature: The DSPy signature to which future LLM responses should conform.
      values: A dictionary mapping field names (from the DSPy signature) to corresponding values
              that should be included in the message.
      role: The role of the message, which can be either "user" or "assistant".
      incomplete: If True, indicates that output field values are present in the set of specified
                  ``values``. If False, indicates that ``values`` only contains input field values.
    Returns:
      A chat message that can be appended to a chat thread. The message contains two string fields:
      ``role`` ("user" or "assistant") and ``content`` (the message text).
    """
    content = []

    if role == "user":
        fields: Dict[str, FieldInfo] = signature.input_fields
        if incomplete:
            content.append("This is an example of the task, though some input or output fields are not supplied.")
    else:
        fields: Dict[str, FieldInfo] = signature.output_fields
        # Add the built-in field indicating that the chat turn has been completed
        fields[BuiltInCompletedOutputFieldInfo.name] = BuiltInCompletedOutputFieldInfo.info
        values = {**values, BuiltInCompletedOutputFieldInfo.name: ""}

    if not incomplete:
        field_names: KeysView = fields.keys()
        if not set(values).issuperset(set(field_names)):
            raise ValueError(f"Expected {field_names} but got {values.keys()}")

    formatted_fields = format_fields(
        fields_with_values={
            FieldInfoWithName(name=field_name, info=field_info): values.get(
                field_name, "Not supplied for this particular example."
            )
            for field_name, field_info in fields.items()
        }
    )
    content.append(formatted_fields)

    if role == "user":
        def type_info(v):
            return f" (must be formatted as a valid Python {get_annotation_name(v.annotation)})" \
                if v.annotation is not str else ""
        
        if not incomplete:
            content.append(
                "Respond with the corresponding output fields, starting with the field "
                + ", then ".join(f"`[[ ## {f} ## ]]`{type_info(v)}" for f, v in signature.output_fields.items())
                + ", and then ending with the marker for `[[ ## completed ## ]]`."
            )

            # content.append(
            #     "Respond with the corresponding output fields, starting with the field "
            #     + ", then ".join(f"`{f}`" for f in signature.output_fields)
            #     + ", and then ending with the marker for `completed`."
            # )

    return {"role": role, "content": "\n\n".join(content).strip()}


def get_annotation_name(annotation):
    origin = get_origin(annotation)
    args = get_args(annotation)
    if origin is None:
        if hasattr(annotation, "__name__"):
            return annotation.__name__
        else:
            return str(annotation)
    else:
        args_str = ", ".join(get_annotation_name(arg) for arg in args)
        return f"{get_annotation_name(origin)}[{args_str}]"


def enumerate_fields(fields):
    parts = []
    for idx, (k, v) in enumerate(fields.items()):
        parts.append(f"{idx+1}. `{k}`")
        parts[-1] += f" ({get_annotation_name(v.annotation)})"
        parts[-1] += f": {v.json_schema_extra['desc']}" if v.json_schema_extra["desc"] != f"${{{k}}}" else ""

    return "\n".join(parts).strip()


def move_type_to_front(d):
    # Move the 'type' key to the front of the dictionary, recursively, for LLM readability/adherence.
    if isinstance(d, Mapping):
        return {k: move_type_to_front(v) for k, v in sorted(d.items(), key=lambda item: (item[0] != 'type', item[0]))}
    elif isinstance(d, list):
        return [move_type_to_front(item) for item in d]
    return d

def prepare_schema(type_):
    schema = pydantic.TypeAdapter(type_).json_schema()
    schema = move_type_to_front(schema)
    return schema

def prepare_instructions(signature: SignatureMeta):
    parts = []
    parts.append("Your input fields are:\n" + enumerate_fields(signature.input_fields))
    parts.append("Your output fields are:\n" + enumerate_fields(signature.output_fields))
    parts.append("All interactions will be structured in the following way, with the appropriate values filled in.")

    def field_metadata(field_name, field_info):
        type_ = field_info.annotation

        if get_dspy_field_type(field_info) == 'input' or type_ is str:
            desc = ""
        elif type_ is bool:
            desc = "must be True or False"
        elif type_ in (int, float):
            desc = f"must be a single {type_.__name__} value"
        elif inspect.isclass(type_) and issubclass(type_, enum.Enum):
            desc= f"must be one of: {'; '.join(type_.__members__)}"
        elif hasattr(type_, '__origin__') and type_.__origin__ is Literal:
            desc = f"must be one of: {'; '.join([str(x) for x in type_.__args__])}"
        else:
            desc = "must be pareseable according to the following JSON schema: "
            desc += json.dumps(prepare_schema(type_))

        desc = (" " * 8) + f"# note: the value you produce {desc}" if desc else ""
        return f"{{{field_name}}}{desc}"

    def format_signature_fields_for_instructions(fields: Dict[str, FieldInfo]):
        return format_fields(
            fields_with_values={
                FieldInfoWithName(name=field_name, info=field_info): field_metadata(field_name, field_info)
                for field_name, field_info in fields.items()
            }
        )

    parts.append(format_signature_fields_for_instructions(signature.input_fields))
    parts.append(format_signature_fields_for_instructions(signature.output_fields))
    parts.append(format_fields({BuiltInCompletedOutputFieldInfo: ""}))

    instructions = textwrap.dedent(signature.instructions)
    objective = ("\n" + " " * 8).join([""] + instructions.splitlines())
    parts.append(f"In adhering to this structure, your objective is: {objective}")

    # parts.append("You will receive some input fields in each interaction. " +
    #              "Respond only with the corresponding output fields, starting with the field " +
    #              ", then ".join(f"`{f}`" for f in signature.output_fields) +
    #              ", and then ending with the marker for `completed`.")

    return "\n\n".join(parts).strip()



import abc
from typing import List, Optional

import numpy as np
import openai
import math
import requests


class BaseSentenceVectorizer(abc.ABC):
    """
    Base Class for Vectorizers. The main purpose is to vectorize text (doc/query)
    for ANN/KNN indexes. `__call__` method takes `List[Example]` as a single input, then extracts
    `field_to_vectorize` from every Example and convert them into embeddings.
    You can customize extraction logic in the `_extract_text_from_examples` method.
    """

    # embeddings will be computed based on the string in this attribute of Example object
    field_to_vectorize = "text_to_vectorize"

    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        pass

    def _extract_text_from_examples(self, inp_examples: List) -> List[str]:
        if isinstance(inp_examples[0], str):
            return inp_examples
        return [" ".join([example[key] for key in example._input_keys]) for example in inp_examples]


class SentenceTransformersVectorizer(BaseSentenceVectorizer):
    """
    Vectorizer based on `SentenceTransformers` models. You can pick any model from this link:
    https://huggingface.co/models?library=sentence-transformers
    More details about models:
    https://www.sbert.net/docs/pretrained_models.html
    """

    def __init__(
        self,
        model_name_or_path: str = "all-MiniLM-L6-v2",
        vectorize_bs: int = 256,
        max_gpu_devices: int = 1,
        normalize_embeddings: bool = False,
    ):
        # this isn't a good practice, but with top-level import the whole DSP
        # module import will be slow (>5 sec), because SentenceTransformer is doing
        # it's directory/file-related magic under the hood :(

        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            raise ImportError(
                "You need to install sentence_transformers library to use pretrained embedders. "
                "Please check the official doc https://www.sbert.net/ "
                "or simply run `pip install sentence-transformers",
            )
        from dsp.utils.ann_utils import determine_devices

        self.num_devices, self.is_gpu = determine_devices(max_gpu_devices)
        self.proxy_device = "cuda" if self.is_gpu else "cpu"

        self.model = SentenceTransformer(model_name_or_path, device=self.proxy_device)

        self.model_name_or_path = model_name_or_path
        self.vectorize_bs = vectorize_bs
        self.normalize_embeddings = normalize_embeddings

    def __call__(self, inp_examples: List) -> np.ndarray:
        text_to_vectorize = self._extract_text_from_examples(inp_examples)

        if self.is_gpu and self.num_devices > 1:
            target_devices = list(range(self.num_devices))
            pool = self.model.start_multi_process_pool(target_devices=target_devices)
            # Compute the embeddings using the multi-process pool
            emb = self.model.encode_multi_process(
                sentences=text_to_vectorize,
                pool=pool,
                batch_size=self.vectorize_bs,
            )
            self.model.stop_multi_process_pool(pool)
            # for some reason, multi-GPU setup doesn't accept normalize_embeddings parameter
            if self.normalize_embeddings:
                emb = emb / np.linalg.norm(emb)

            return emb
        else:
            emb = self.model.encode(
                sentences=text_to_vectorize,
                batch_size=self.vectorize_bs,
                normalize_embeddings=self.normalize_embeddings,
            )
            return emb


class NaiveGetFieldVectorizer(BaseSentenceVectorizer):
    """
    If embeddings were precomputed, then we could just extract them from the proper field
    (set by `field_with_embedding`) from each `Example`.
    """

    def __init__(self, field_with_embedding: str = "vectorized"):
        self.field_with_embedding = field_with_embedding

    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        embeddings = [getattr(cur_example, self.field_with_embedding).reshape(1, -1) for cur_example in inp_examples]
        embeddings = np.concatenate(embeddings, axis=0).astype(np.float32)
        return embeddings


class CohereVectorizer(BaseSentenceVectorizer):
    """
    This vectorizer uses the Cohere API to convert texts to embeddings.
    More about the available models: https://docs.cohere.com/reference/embed
    `api_key` should be passed as an argument and can be retrieved
    from https://dashboard.cohere.com/api-keys
    """

    def __init__(
        self,
        api_key: str,
        model: str = "embed-english-v3.0",
        embed_batch_size: int = 96,
        embedding_type: str = "search_document",  # for details check Cohere embed docs
    ):
        self.model = model
        self.embed_batch_size = embed_batch_size
        self.embedding_type = embedding_type

        import cohere

        self.client = cohere.Client(api_key, client_name="dspy")

    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        text_to_vectorize = self._extract_text_from_examples(inp_examples)

        embeddings_list = []

        n_batches = (len(text_to_vectorize) - 1) // self.embed_batch_size + 1
        for cur_batch_idx in range(n_batches):
            start_idx = cur_batch_idx * self.embed_batch_size
            end_idx = (cur_batch_idx + 1) * self.embed_batch_size
            cur_batch = text_to_vectorize[start_idx:end_idx]

            response = self.client.embed(
                texts=cur_batch,
                model=self.model,
                input_type=self.embedding_type,
            )

            embeddings_list.extend(response.embeddings)

        embeddings = np.array(embeddings_list, dtype=np.float32)
        return embeddings


try:
    OPENAI_LEGACY = int(openai.version.__version__[0]) == 0
except Exception:
    OPENAI_LEGACY = True


class OpenAIVectorizer(BaseSentenceVectorizer):
    """
    This vectorizer uses OpenAI API to convert texts to embeddings. Changing `model` is not
    recommended. More about the model: https://openai.com/blog/new-and-improved-embedding-model/
    `api_key` should be passed as an argument or as env variable (`OPENAI_API_KEY`).
    """

    def __init__(
        self,
        model: str = "text-embedding-ada-002",
        embed_batch_size: int = 1024,
        api_key: Optional[str] = None,
    ):
        self.model = model
        self.embed_batch_size = embed_batch_size

        if OPENAI_LEGACY:
            self.Embedding = openai.Embedding
        else:
            self.Embedding = openai.embeddings

        if api_key:
            openai.api_key = api_key

    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        text_to_vectorize = self._extract_text_from_examples(inp_examples)
        # maybe it's better to preallocate numpy matrix, but we don't know emb_dim
        embeddings_list = []

        n_batches = (len(text_to_vectorize) - 1) // self.embed_batch_size + 1
        for cur_batch_idx in range(n_batches):  # tqdm.tqdm?
            start_idx = cur_batch_idx * self.embed_batch_size
            end_idx = (cur_batch_idx + 1) * self.embed_batch_size
            cur_batch = text_to_vectorize[start_idx:end_idx]
            # OpenAI API call:
            response = self.Embedding.create(
                model=self.model,
                input=cur_batch,
            )

            cur_batch_embeddings = [cur_obj["embedding"] for cur_obj in response["data"]]
            embeddings_list.extend(cur_batch_embeddings)

        embeddings = np.array(embeddings_list, dtype=np.float32)
        return embeddings


class FastEmbedVectorizer(BaseSentenceVectorizer):
    """Sentence vectorizer implementaion using FastEmbed - https://qdrant.github.io/fastembed."""

    def __init__(
        self,
        model_name: str = "BAAI/bge-small-en-v1.5",
        batch_size: int = 256,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        parallel: Optional[int] = None,
        **kwargs,
    ):
        """Initialize fastembed.TextEmbedding.

        Args:
            model_name (str): The name of the model to use. Defaults to `"BAAI/bge-small-en-v1.5"`.
            batch_size (int): Batch size for encoding. Higher values will use more memory, but be faster.\
                                        Defaults to 256.
            cache_dir (str, optional): The path to the model cache directory.\
                                       Can also be set using the `FASTEMBED_CACHE_PATH` env variable.
            threads (int, optional): The number of threads single onnxruntime session can use.
            parallel (int, optional): If `>1`, data-parallel encoding will be used, recommended for large datasets.\
                                      If `0`, use all available cores.\
                                      If `None`, don't use data-parallel processing, use default onnxruntime threading.\
                                      Defaults to None.
            **kwargs: Additional options to pass to fastembed.TextEmbedding
        Raises:
            ValueError: If the model_name is not in the format <org>/<model> e.g. BAAI/bge-small-en-v1.5.
        """
        try:
            from fastembed import TextEmbedding
        except ImportError as e:
            raise ValueError(
                "The 'fastembed' package is not installed. Please install it with `pip install fastembed`",
            ) from e
        self._batch_size = batch_size
        self._parallel = parallel
        self._model = TextEmbedding(model_name=model_name, cache_dir=cache_dir, threads=threads, **kwargs)

    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        texts_to_vectorize = self._extract_text_from_examples(inp_examples)
        embeddings = self._model.embed(texts_to_vectorize, batch_size=self._batch_size, parallel=self._parallel)

        return np.array([embedding.tolist() for embedding in embeddings], dtype=np.float32)


class PremAIVectorizer(BaseSentenceVectorizer):
    """The PremAIVectorizer class utilizes the PremAI Embeddings API to convert text into embeddings.
    This vectorizer leverages various models provided by PremAI.

    For detailed information on the supported models, visit: https://docs.premai.io/get-started/supported-models.

    The `project_id` is a mandatory argument, while `api_key` and `model_name` are optional. The `api_key`
    can be supplied either as an argument or through an environment variable. By default, the `model_name`
    is set to "text-embedding-3-large", unless specified otherwise.

    To learn more about getting started with PremAI, visit: https://docs.premai.io/introduction.
    """

    def __init__(
        self,
        project_id: str,
        api_key: Optional[str] = None,
        model_name: Optional[str] = "text-embedding-3-large",
        embed_batch_size: int = 32,
    ):
        self.model_name, self.project_id = model_name, project_id
        self.embed_batch_size = embed_batch_size

        try:
            from premai import Prem

            from dsp.modules.premai import get_premai_api_key

            api_key = get_premai_api_key(api_key=api_key)
            self.client = Prem(api_key=api_key)
        except ImportError as error:
            raise ImportError("Please install premai package using: pip install premai") from error

    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        text_to_vectorize = self._extract_text_from_examples(inp_examples)
        embedding_list = []

        n_batches = (len(text_to_vectorize) - 1) // self.embed_batch_size + 1
        for cur_batch_idx in range(n_batches):
            start_idx = cur_batch_idx * self.embed_batch_size
            end_idx = (cur_batch_idx + 1) * self.embed_batch_size
            current_batch = text_to_vectorize[start_idx:end_idx]
            embeddings = self.client.embeddings.create(
                project_id=self.project_id,
                model=self.model_name,
                input=current_batch,
            ).data
            current_batch_embeddings = [embedding.embedding for embedding in embeddings]
            embedding_list.extend(current_batch_embeddings)

        embeddings = np.array(embedding_list, dtype=np.float32)
        return embeddings


class TEIVectorizer(BaseSentenceVectorizer):
    """The TEIVectorizer class utilizes the TEI(Text Embeddings Inference) Embeddings API to
    convert text into embeddings.

    For detailed information on the supported models, visit: https://github.com/huggingface/text-embeddings-inference.

    `model` is embedding model name.
    `embed_batch_size` is the maximum batch size for a single request.
    `api_key` request authorization.
    `api_url` custom inference endpoint url.

    To learn more about getting started with TEI, visit: https://github.com/huggingface/text-embeddings-inference.
    """

    def __init__(
        self,
        model: Optional[str] = "bge-base-en-v1.5",
        embed_batch_size: int = 256,
        api_key: Optional[str] = None,
        api_url: str = "",
    ):
        self.model = model
        self.embed_batch_size = embed_batch_size
        self.api_key = api_key
        self.api_url = api_url

    @property
    def _headers(self) -> dict:
        return {"Authorization": f"Bearer {self.api_key}"}

    def __call__(self, inp_examples: List["Example"]) -> np.ndarray:
        text_to_vectorize = self._extract_text_from_examples(inp_examples)
        embeddings_list = []

        n = math.ceil(len(text_to_vectorize) / self.embed_batch_size)
        for i in range(n):
            response = requests.post(
                self.api_url,
                headers=self._headers,
                json={
                    "inputs": text_to_vectorize[i * self.embed_batch_size:(i + 1) * self.embed_batch_size],
                    "normalize": True,
                    "truncate": True
                },
            )
            embeddings_list.extend(response.json())

        embeddings = np.array(embeddings_list, dtype=np.float32)
        return embeddings



"""AWS providers for LMs."""

from abc import ABC, abstractmethod
from typing import Any, Optional

import backoff
from dsp.utils.settings import settings

try:
    import boto3
    from botocore.exceptions import ClientError
    ERRORS = (ClientError,)

except ImportError:
    ERRORS = (Exception,)


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/."""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


def giveup_hdlr(details):
    """Wrapper function that decides when to give up on retry."""
    if "max retries" in details.args[0]:
        return False
    return True

class AWSProvider(ABC):
    """This abstract class adds support for AWS model providers such as Bedrock and SageMaker.
    The subclasses such as Bedrock and Sagemaker implement the abstract method _call_model and work in conjunction with the AWSModel classes.
    Usage Example:
        bedrock = dspy.Bedrock(region_name="us-west-2")
        bedrock_mixtral = dspy.AWSMistral(bedrock, "mistral.mixtral-8x7b-instruct-v0:1", **kwargs)
        bedrock_haiku = dspy.AWSAnthropic(bedrock, "anthropic.claude-3-haiku-20240307-v1:0", **kwargs)
        bedrock_llama2 = dspy.AWSMeta(bedrock, "meta.llama2-13b-chat-v1", **kwargs)

        sagemaker = dspy.Sagemaker(region_name="us-west-2")
        sagemaker_model = dspy.AWSMistral(sagemaker, "<YOUR_ENDPOINT_NAME>", **kwargs)
    """

    def __init__(
        self,
        region_name: str,
        service_name: str,
        profile_name: Optional[str] = None,
        batch_n_enabled: bool = True,
    ) -> None:
        """_summary_.

        Args:
            region_name (str, optional): The AWS region where this LM is hosted.
            service_name (str): Used in context of invoking the boto3 API.
            profile_name (str, optional): boto3 credentials profile.
            batch_n_enabled (bool): If False, call the LM N times rather than batching.
        """
        try:
            import boto3
        except ImportError as exc:
            raise ImportError('pip install boto3 to use AWS models.') from exc

        if profile_name is None:
            self.predictor = boto3.client(service_name, region_name=region_name)
        else:
            self.predictor = boto3.Session(profile_name=profile_name).client(
                service_name,
                region_name=region_name,
            )

        self.batch_n_enabled = batch_n_enabled

    def get_provider_name(self) -> str:
        """Return the provider name."""
        return self.__class__.__name__

    @abstractmethod
    @backoff.on_exception(
        backoff.expo,
        ERRORS,
        max_time=settings.backoff_time,
        max_tries=8,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def call_model(self, model_id: str, body: str) -> str:
        """Call the model and return the response."""

    def sanitize_kwargs(self, query_kwargs: dict[str, Any]) -> tuple[int, dict[str, Any]]:
        """Ensure that input kwargs can be used by Bedrock or Sagemaker."""
        if "temperature" in query_kwargs:
            if query_kwargs["temperature"] > 0.99:
                query_kwargs["temperature"] = 0.99
            if query_kwargs["temperature"] < 0.01:
                query_kwargs["temperature"] = 0.01

        if "top_p" in query_kwargs:
            if query_kwargs["top_p"] > 0.99:
                query_kwargs["top_p"] = 0.99
            if query_kwargs["top_p"] < 0.01:
                query_kwargs["top_p"] = 0.01

        n = -1
        if not self.batch_n_enabled:
            n = query_kwargs.pop('n', 1)
            query_kwargs["num_generations"] = n

        return n, query_kwargs


class Bedrock(AWSProvider):
    """This class adds support for Bedrock models."""

    def __init__(
        self,
        region_name: str,
        profile_name: Optional[str] = None,
        batch_n_enabled: bool = False,   # This has to be setup manually on Bedrock.
    ) -> None:
        """_summary_.

        Args:
            region_name (str): The AWS region where this LM is hosted.
            profile_name (str, optional): boto3 credentials profile.
        """
        super().__init__(region_name, "bedrock-runtime", profile_name, batch_n_enabled)

    def call_model(self, model_id: str, body: str) -> str:
        return self.predictor.invoke_model(
            modelId=model_id,
            body=body,
            accept="application/json",
            contentType="application/json",
        )


class Sagemaker(AWSProvider):
    """This class adds support for Sagemaker models."""

    def __init__(
        self,
        region_name: str,
        profile_name: Optional[str] = None,
    ) -> None:
        """_summary_.

        Args:
            region_name (str, optional): The AWS region where this LM is hosted.
            profile_name (str, optional): boto3 credentials profile.
        """
        super().__init__(region_name, "runtime.sagemaker", profile_name)

    @backoff.on_exception(
        backoff.expo,
        ERRORS,
        max_time=settings.backoff_time,
        max_tries=8,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def call_model(self, model_id: str, body: str) -> str:
        return self.predictor.invoke_endpoint(
            EndpointName=model_id,
            Body=body,
            Accept="application/json",
            ContentType="application/json",
        )



import types
from typing import List, Optional

import dsp
from dspy.predict.knn import KNN
from dspy.teleprompt import BootstrapFewShot

from .teleprompt import Teleprompter


class KNNFewShot(Teleprompter):
    def __init__(self, k: int, trainset: List[dsp.Example], vectorizer: Optional[dsp.BaseSentenceVectorizer] = None, **few_shot_bootstrap_args):
        self.KNN = KNN(k, trainset, vectorizer=vectorizer)
        self.few_shot_bootstrap_args = few_shot_bootstrap_args

    def compile(self, student, *, teacher=None, trainset=None, valset=None):
        student_copy = student.reset_copy()

        def forward_pass(_, **kwargs):
            knn_trainset = self.KNN(**kwargs)
            few_shot_bootstrap = BootstrapFewShot(**self.few_shot_bootstrap_args)
            compiled_program = few_shot_bootstrap.compile(
                student,
                teacher=teacher,
                trainset=knn_trainset,
            )
            return compiled_program(**kwargs)

        student_copy.forward = types.MethodType(forward_pass, student_copy)
        return student_copy



import functools
import json
import logging
from typing import Any, Callable, Literal, Optional, cast

import backoff

try:
    """
    If there is any error in the langfuse configuration, it will turn to request the real address(openai or azure endpoint)
    """
    import langfuse
    from langfuse.openai import openai
    logging.info(f"You are using Langfuse,version{langfuse.__version__}")
except:
    import openai

from dsp.modules.cache_utils import CacheMemory, NotebookCacheMemory, cache_turn_on
from dsp.modules.lm import LM
from dsp.utils.settings import settings


try:
    OPENAI_LEGACY = int(openai.version.__version__[0]) == 0
except Exception:
    OPENAI_LEGACY = True

try:
    import openai.error
    from openai.openai_object import OpenAIObject

    ERRORS = (
        openai.error.RateLimitError,
        openai.error.ServiceUnavailableError,
        openai.error.APIError,
    )
except Exception:
    ERRORS = (openai.RateLimitError, openai.APIError)
    OpenAIObject = dict


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


AzureADTokenProvider = Callable[[], str]


class AzureOpenAI(LM):
    """Wrapper around Azure's API for OpenAI.

    Args:
        api_base (str): Azure URL endpoint for model calling, often called 'azure_endpoint'.
        api_version (str): Version identifier for API.
        model (str, optional): OpenAI or Azure supported LLM model to use. Defaults to "gpt-3.5-turbo-instruct".
        api_key (Optional[str], optional): API provider Authentication token. use Defaults to None.
        model_type (Literal["chat", "text"], optional): The type of model that was specified. Mainly to decide the optimal prompting strategy. Defaults to "chat".
        **kwargs: Additional arguments to pass to the API provider.
    """

    def __init__(
        self,
        api_base: str,
        api_version: str,
        model: str = "gpt-3.5-turbo-instruct",
        api_key: Optional[str] = None,
        model_type: Literal["chat", "text"] = "chat",
        system_prompt: Optional[str] = None,
        azure_ad_token_provider: Optional[AzureADTokenProvider] = None,
        **kwargs,
    ):
        super().__init__(model)
        self.provider = "openai"

        self.system_prompt = system_prompt

        # Define Client
        if OPENAI_LEGACY:
            # Assert that all variables are available
            assert (
                "engine" in kwargs or "deployment_id" in kwargs
            ), "Must specify engine or deployment_id for Azure API instead of model."

            openai.api_base = api_base
            openai.api_key = api_key
            openai.api_type = "azure"
            openai.api_version = api_version
            openai.azure_ad_token_provider = azure_ad_token_provider

            self.client = None

        else:
            client = openai.AzureOpenAI(
                azure_endpoint=api_base,
                api_key=api_key,
                api_version=api_version,
                azure_ad_token_provider=azure_ad_token_provider,
            )

            self.client = client

        self.model_type = model_type

        if not OPENAI_LEGACY and "model" not in kwargs:
            if "deployment_id" in kwargs:
                kwargs["model"] = kwargs["deployment_id"]
                del kwargs["deployment_id"]

            if "api_version" in kwargs:
                del kwargs["api_version"]

        if "model" not in kwargs:
            kwargs["model"] = model

        self.kwargs = {
            "temperature": 0.0,
            "max_tokens": 150,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "n": 1,
            **kwargs,
        }  # TODO: add kwargs above for </s>

        self.api_base = api_base
        self.api_version = api_version
        self.api_key = api_key

        self.history: list[dict[str, Any]] = []

    def _openai_client(self):
        if OPENAI_LEGACY:
            return openai

        return self.client

    def log_usage(self, response):
        """Log the total tokens from the Azure OpenAI API response."""
        usage_data = response.get("usage")
        if usage_data:
            total_tokens = usage_data.get("total_tokens")
            logging.debug(f"Azure OpenAI Total Token Usage: {total_tokens}")

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs

        kwargs = {**self.kwargs, **kwargs}
        if self.model_type == "chat":
            # caching mechanism requires hashable kwargs
            messages = [{"role": "user", "content": prompt}]
            if self.system_prompt:
                messages.insert(0, {"role": "system", "content": self.system_prompt})

            kwargs["messages"] = messages
            kwargs = {"stringify_request": json.dumps(kwargs)}
            response = chat_request(self.client, **kwargs)

        else:
            kwargs["prompt"] = prompt
            response = completions_request(self.client, **kwargs)

        history = {
            "prompt": prompt,
            "response": response,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)

        return response

    @backoff.on_exception(
        backoff.expo,
        ERRORS,
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of GPT-3 completions whilst handling rate limiting and caching."""
        if "model_type" in kwargs:
            del kwargs["model_type"]

        return self.basic_request(prompt, **kwargs)

    def _get_choice_text(self, choice: dict[str, Any]) -> str:
        if self.model_type == "chat":
            return choice["message"]["content"]
        return choice["text"]

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ) -> list[dict[str, Any]]:
        """Retrieves completions from OpenAI Model.

        Args:
            prompt (str): prompt to send to GPT-3
            only_completed (bool, optional): return only completed responses and ignores completion due to length. Defaults to True.
            return_sorted (bool, optional): sort the completion choices using the returned probabilities. Defaults to False.

        Returns:
            list[dict[str, Any]]: list of completion choices
        """

        assert only_completed, "for now"
        assert return_sorted is False, "for now"

        response = self.request(prompt, **kwargs)

        self.log_usage(response)

        choices = response["choices"]

        completed_choices = [c for c in choices if c["finish_reason"] != "length"]

        if only_completed and len(completed_choices):
            choices = completed_choices

        completions = [self._get_choice_text(c) for c in choices]
        if return_sorted and kwargs.get("n", 1) > 1:
            scored_completions = []

            for c in choices:
                tokens, logprobs = (
                    c["logprobs"]["tokens"],
                    c["logprobs"]["token_logprobs"],
                )

                if "<|endoftext|>" in tokens:
                    index = tokens.index("<|endoftext|>") + 1
                    tokens, logprobs = tokens[:index], logprobs[:index]

                avglog = sum(logprobs) / len(logprobs)
                scored_completions.append((avglog, self._get_choice_text(c)))

            scored_completions = sorted(scored_completions, reverse=True)
            completions = [c for _, c in scored_completions]

        return completions

    def copy(self, **kwargs):
        """Returns a copy of the language model with the same parameters."""
        kwargs = {**self.kwargs, **kwargs}
        model = kwargs.pop("model")

        return self.__class__(
            model=model,
            api_key=self.api_key,
            api_version=self.api_version,
            api_base=self.api_base,
            **kwargs,
        )


@CacheMemory.cache
def cached_gpt3_request_v2(**kwargs):
    return openai.Completion.create(**kwargs)


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def cached_gpt3_request_v2_wrapped(**kwargs):
    return cached_gpt3_request_v2(**kwargs)


@CacheMemory.cache
def _cached_gpt3_turbo_request_v2(**kwargs) -> OpenAIObject:
    if "stringify_request" in kwargs:
        kwargs = json.loads(kwargs["stringify_request"])
    return cast(OpenAIObject, openai.ChatCompletion.create(**kwargs))


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def _cached_gpt3_turbo_request_v2_wrapped(**kwargs) -> OpenAIObject:
    return _cached_gpt3_turbo_request_v2(**kwargs)


def v1_chat_request(client, **kwargs):
    @functools.lru_cache(maxsize=None if cache_turn_on else 0)
    @NotebookCacheMemory.cache
    def v1_cached_gpt3_turbo_request_v2_wrapped(**kwargs):
        @CacheMemory.cache
        def v1_cached_gpt3_turbo_request_v2(**kwargs):
            if "stringify_request" in kwargs:
                kwargs = json.loads(kwargs["stringify_request"])
            return client.chat.completions.create(**kwargs)

        return v1_cached_gpt3_turbo_request_v2(**kwargs)

    return v1_cached_gpt3_turbo_request_v2_wrapped(**kwargs).model_dump()


def v1_completions_request(client, **kwargs):
    @functools.lru_cache(maxsize=None if cache_turn_on else 0)
    @NotebookCacheMemory.cache
    def v1_cached_gpt3_request_v2_wrapped(**kwargs):
        @CacheMemory.cache
        def v1_cached_gpt3_request_v2(**kwargs):
            return client.completions.create(**kwargs)

        return v1_cached_gpt3_request_v2(**kwargs)

    return v1_cached_gpt3_request_v2_wrapped(**kwargs).model_dump()


def chat_request(client, **kwargs):
    if OPENAI_LEGACY:
        return _cached_gpt3_turbo_request_v2_wrapped(**kwargs)

    return v1_chat_request(client, **kwargs)


def completions_request(client, **kwargs):
    if OPENAI_LEGACY:
        return cached_gpt3_request_v2_wrapped(**kwargs)

    return v1_completions_request(client, **kwargs)



import logging
from collections.abc import Iterable

import numpy as np

import dsp

logger = logging.getLogger(__name__)

def retrieve(query: str, k: int, **kwargs) -> list[str]:
    """Retrieves passages from the RM for the query and returns the top k passages."""
    if not dsp.settings.rm:
        raise AssertionError("No RM is loaded.")
    passages = dsp.settings.rm(query, k=k, **kwargs)
    if not isinstance(passages, Iterable):
        # it's not an iterable yet; make it one.
        # TODO: we should unify the type signatures of dspy.Retriever
        passages = [passages]
    passages = [psg.long_text for psg in passages]

    if dsp.settings.reranker:
        passages_cs_scores = dsp.settings.reranker(query, passages)
        passages_cs_scores_sorted = np.argsort(passages_cs_scores)[::-1]
        passages = [passages[idx] for idx in passages_cs_scores_sorted]


    return passages
def retrievewithMetadata(query: str, k: int, **kwargs) -> list[str]:
    """Retrieves passages from the RM for the query and returns the top k passages."""

    if not dsp.settings.rm:
        raise AssertionError("No RM is loaded.")
    passages = dsp.settings.rm(query, k=k, **kwargs)
    if not isinstance(passages, Iterable):
        # it's not an iterable yet; make it one.
        # TODO: we should unify the type signatures of dspy.Retriever
        passages = [passages]

    return passages


def retrieveRerankEnsemble(queries: list[str], k: int,**kwargs) -> list[str]:
    if not (dsp.settings.rm and dsp.settings.reranker):
        raise AssertionError("Both RM and Reranker are needed to retrieve & re-rank.")
    queries = [q for q in queries if q]
    passages = {}
    for query in queries:
        retrieved_passages = dsp.settings.rm(query, k=k*3,**kwargs)
        passages_cs_scores = dsp.settings.reranker(query, [psg.long_text for psg in retrieved_passages])
        for idx in np.argsort(passages_cs_scores)[::-1]:
            psg = retrieved_passages[idx]
            passages[psg.long_text] = passages.get(psg.long_text, []) + [
                passages_cs_scores[idx],
            ]


    passages = [(np.average(score), text) for text, score in passages.items()]
    return [text for _, text in sorted(passages, reverse=True)[:k]]

def retrieveRerankEnsemblewithMetadata(queries: list[str], k: int, **kwargs) -> list[str]:
    if not (dsp.settings.rm and dsp.settings.reranker):
        raise AssertionError("Both RM and Reranker are needed to retrieve & re-rank.")
    queries = [q for q in queries if q]
    all_queries_passages = []
    for query in queries:
        passages = []
        retrieved_passages = dsp.settings.rm(query, k=k * 3, **kwargs)
        passages_cs_scores = dsp.settings.reranker(
            query, passages=[psg["long_text"] for psg in retrieved_passages],
        )
        for idx in np.argsort(passages_cs_scores)[::-1][:k]:
            curr_passage = retrieved_passages[idx]
            curr_passage["rerank_score"] = passages_cs_scores[idx]
            passages.append(curr_passage)
        all_queries_passages.append(passages)
    if len(queries) == 1:
        return all_queries_passages[0]
    else:
        return all_queries_passages


def retrieveEnsemble(queries: list[str], k: int, by_prob: bool = True,**kwargs) -> list[str]:
    """Retrieves passages from the RM for each query in queries and returns the top k passages
    based on the probability or score.
    """
    if not dsp.settings.rm:
        raise AssertionError("No RM is loaded.")
    if dsp.settings.reranker:
        return retrieveRerankEnsemble(queries, k, **kwargs)
    queries = [q for q in queries if q]

    if len(queries) == 1:
        return retrieve(queries[0], k, **kwargs)

    passages = {}
    for q in queries:
        for psg in dsp.settings.rm(q, k=k * 3,**kwargs):
            if by_prob:
                passages[psg.long_text] = passages.get(psg.long_text, 0.0) + psg.prob
            else:
                passages[psg.long_text] = passages.get(psg.long_text, 0.0) + psg.score

    passages = [(score, text) for text, score in passages.items()]
    passages = sorted(passages, reverse=True)[:k]
    passages = [text for _, text in passages]


    return passages

def retrieveEnsemblewithMetadata(
    queries: list[str], k: int, by_prob: bool = True, **kwargs,
) -> list[str]:
    """Retrieves passages from the RM for each query in queries and returns the top k passages
    based on the probability or score.
    """

    if not dsp.settings.rm:
        raise AssertionError("No RM is loaded.")
    if not dsp.settings.reranker:
        return retrieveRerankEnsemblewithMetadata(queries=queries,k=k)

    queries = [q for q in queries if q]

    if len(queries) == 1:
        return retrieve(queries[0], k)
    all_queries_passages = []
    for q in queries:
        passages = {}
        retrieved_passages = dsp.settings.rm(q, k=k * 3, **kwargs)
        for idx, psg in enumerate(retrieved_passages):
            if by_prob:
                passages[(idx, psg.long_text)] = (
                    passages.get(psg.long_text, 0.0) + psg.prob
                )
            else:
                passages[(idx, psg.long_text)] = (
                    passages.get(psg.long_text, 0.0) + psg.score
                )
            retrieved_passages[idx]["tracking_idx"] = idx
        passages = sorted(passages.items(), key=lambda item: item[1])[:k]
        req_indices = [psg[0][0] for psg in passages]
        passages = [
            rp for rp in retrieved_passages if rp.get("tracking_idx") in req_indices
        ]
        all_queries_passages.append(passages)
    return all_queries_passages